from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from fileinput import filename

import os
import os.path as osp
import math
import time
import glob
import abc
from antgo.framework.helper.runner.hooks import HOOKS, Hook
from antgo.framework.helper.utils import Registry, build_from_cfg
from torch.utils.data import DataLoader
import torch.optim
import torchvision.transforms as transforms
from torch.nn.parallel.data_parallel import DataParallel
from antgo.framework.helper.utils.config import Config
from antgo.framework.helper.apis.train import *
from antgo.framework.helper.dataset import (build_dataloader,build_kv_dataloader, build_dataset)
from antgo.framework.helper.utils.util_distribution import build_ddp, build_dp, get_device
from antgo.framework.helper.utils import get_logger
from antgo.framework.helper.runner import get_dist_info, init_dist
from antgo.framework.helper.runner.builder import *
from antgo.framework.helper.models.builder import *
import torch.distributed as dist
from contextlib import contextmanager
from antgo.framework.helper.utils.setup_env import *
from antgo.framework.helper.runner.checkpoint import load_checkpoint
from antgo.framework.helper.runner.test import multi_gpu_test, single_gpu_test
from antgo.framework.helper.cnn.utils import fuse_conv_bn
from antgo.framework.helper.tester import Tester
from antgo.framework.helper.runner.hooks.samplingmethods.kcenter_greedy import *
from antgo.framework.helper.runner import BaseModule
from thop import profile
from antgo.utils.sample_gt import *
from antgo.framework.helper.dataset.builder import DATASETS
from antgo.framework.helper.task_flag import *
import json
import imagesize
from antgo.framework.helper.fileio import *


UNCERTAINTY_SAMPLING = Registry('UNCERTAINTY_SAMPLING')

def softmax(x, axis=None):
  x = x - x.max(axis=axis, keepdims=True)
  y = np.exp(x)
  return y / y.sum(axis=axis, keepdims=True)


@UNCERTAINTY_SAMPLING.register_module()
class SamplingByEntropy(object):
    def __init__(self, key, sampling_num=0, use_softmax=True, ratio=0.1) -> None:
        super().__init__()
        self.key = key
        self.sampling_num = sampling_num
        self.use_softmax = use_softmax
        self.ratio = ratio
    
    def __call__(self, results, sampling_num, pre_selected=None, ref_results=None):
        entropy_list = []
        if pre_selected is None:
            pre_selected = list(range(len(results))) 
        
        if self.sampling_num > 0:
            sampling_num = self.sampling_num
        for i in pre_selected:
            # hand_class_preds: NxBx4
            hand_class_preds = np.stack(results[i][self.key], 0)
            # multi-round mean
            hand_class_preds = np.mean(hand_class_preds, 0)            
            if self.use_softmax:
                hand_class_preds = softmax(hand_class_preds, -1)

            C = hand_class_preds.shape[-1]
            hand_class_preds = np.reshape(hand_class_preds, (-1, C))
            entropy_val_list = []
            for j in range(hand_class_preds.shape[0]):
                entropy_val_list.append(
                    entropy(hand_class_preds[j])
                )
     
            if self.ratio > 0:
                entropy_val_list = sorted(entropy_val_list, reverse=True)                       
                use_num = (int)(self.ratio * len(entropy_val_list))
                entropy_val_list = entropy_val_list[:use_num]
            entropy_list.append((np.mean(entropy_val_list), i))

        ordered_unlabeled = sorted(entropy_list, key=lambda x: x[0], reverse=True)
        ordered_unlabeled = ordered_unlabeled[:sampling_num]
        ordered_indexs = [s[1] for s in ordered_unlabeled]
        
        sample_list = []
        selected_list = []
        for index in ordered_indexs:
            # get index in dataset
            selected_list.append(index)

            meta_info = results[index]['image_meta'][0]
            image_file = meta_info['image_file']
            image_index = meta_info.get('id', '')
            sample_list.append({
                'id': image_index,
                'image_file': image_file,
                'tag': meta_info['tag']
            })

        return sample_list, selected_list
        

@UNCERTAINTY_SAMPLING.register_module()
class SamplingByUniform(object):
    def __init__(self, sampling_num=0) -> None:
        super().__init__()
        self.sampling_num = sampling_num
    
    def __call__(self, results, sampling_num, pre_selected=None, ref_results=None):
        # 1.step uniform sampling
        if pre_selected is None:
            pre_selected = list(range(len(results)))        

        if self.sampling_num > 0:
            sampling_num = self.sampling_num          
        selected_i_list = list(range(len(pre_selected)))
        np.random.shuffle(selected_i_list)
        selected_i_list = selected_i_list[:sampling_num]

        # 2.step select
        sample_list = []
        selected_list = []
        for index in selected_i_list:
            # get index in dataset
            index = pre_selected[index]
            selected_list.append(index)      

            meta_info = results[index]['image_meta'][0]
            image_file = meta_info['image_file']
            image_index = meta_info.get('id', '')
            sample_list.append({
                'id': image_index,
                'image_file': image_file,
                'tag': meta_info['tag']
            })

        return sample_list, selected_list


@UNCERTAINTY_SAMPLING.register_module()
class SamplingByKCenter(object):
    def __init__(self, key, sampling_num=0) -> None:
        super().__init__()
        self.key = key
        self.sampling_num = sampling_num
    
    def __call__(self, results, sampling_num, pre_selected=None, ref_results=None):
        # 1.step stack feature
        feature_stack = []
        if pre_selected is None:
            pre_selected = list(range(len(results)))        
        for i in pre_selected:
            feature = results[i][self.key]
            if type(feature) == list:
                feature = np.mean(np.stack(feature, 0),0)

            feature_stack.append(feature.flatten())
        # NxC
        feature_stack = np.stack(feature_stack, 0)

        # 2.step centergreedy
        if self.sampling_num > 0:
            sampling_num = self.sampling_num        
        kcenter_greedy = kCenterGreedy(feature_stack)
        selected_i_list = kcenter_greedy.select_batch(already_selected=[], N=sampling_num)

        # 3.step select sample
        sample_list = []
        selected_list = []
        for index in selected_i_list:
            # get index in dataset
            index = pre_selected[index]
            selected_list.append(index)    

            meta_info = results[index]['image_meta'][0]
            image_file = meta_info['image_file']
            image_index = meta_info.get('id', '')
            sample_list.append({
                'id': image_index,
                'image_file': image_file,
                'tag': meta_info['tag']
            })

        return sample_list, selected_list


@UNCERTAINTY_SAMPLING.register_module()
class GuidByCrossDomain(object):
    def __init__(self, key, sampling_num=0) -> None:
        super().__init__()
        self.key = key
        self.sampling_num = sampling_num
            
    def __call__(self, results, sampling_num, pre_selected=None, ref_results=None):
        # 1.step stack feature
        feature_stack = []
        if pre_selected is None:
            pre_selected = list(range(len(results)))        
        for i in pre_selected:
            feature = results[i][self.key]
            if type(feature) == list:
                feature = np.mean(np.stack(feature, 0),0)

            feature_stack.append(feature.flatten())
        # NxC
        feature_stack = np.stack(feature_stack, 0)
        N = feature_stack.shape[0]

        # 2.step stack ref feature
        ref_feature_stack = []
        for i in range(len(ref_results)):
            ref_feature = ref_results[i][self.key]
            if type(ref_feature) == list:
                ref_feature = np.mean(np.stack(ref_feature, 0),0)

            ref_feature_stack.append(ref_feature.flatten())
        # NxC
        ref_feature_stack = np.stack(ref_feature_stack, 0)

        # 3.step mining
        if self.sampling_num > 0:
            sampling_num = self.sampling_num        
        
        min_distances = None
        new_batch = []
        ind = 0
        inf_distances = np.zeros((feature_stack.shape[0], 1))
        ref_ind_list = []
        for _ in range(sampling_num):
            if min_distances is None:
                ind = np.random.choice(np.arange(N))
                inf_distances[ind] = 10000000.0
                min_distances = pairwise_distances(ref_feature_stack, feature_stack[ind:ind+1])
            else:
                ref_ind = np.argmax(min_distances)
                temp = pairwise_distances(feature_stack, ref_feature_stack[ref_ind:ref_ind+1])
                temp = np.maximum(inf_distances, temp)
                ind = np.argmin(temp)
                inf_distances[ind] = 10000000.0
            
                # 计算训练集中的ind样本，与参考集中的样本的最小距离，并更新
                min_distances_new = pairwise_distances(ref_feature_stack, feature_stack[ind:ind+1])
                min_distances = np.minimum(min_distances, min_distances_new)
                min_distances[ref_ind] = 0.0
                ref_ind_list.append(ref_ind)
            
            new_batch.append(ind)

        # 4.step select sample
        sample_list = []
        selected_list = []
        for index in new_batch:
            # get index in dataset
            index = pre_selected[index]
            selected_list.append(index)

            meta_info = results[index]['image_meta'][0]
            image_file = meta_info['image_file']
            image_index = meta_info.get('id', '')
            sample_list.append({
                'id': image_index,
                'image_file': image_file,
                'tag': meta_info['tag']
            })

        return sample_list, selected_list


@UNCERTAINTY_SAMPLING.register_module()
class SmplingByMix(object):
    def __init__(self, sampling_cfgs):
        super().__init__()
        self.sampling_fns= []
        for cfg in sampling_cfgs:
            sampling_fn = build_from_cfg(cfg, UNCERTAINTY_SAMPLING)
            self.sampling_fns.append(sampling_fn)        

    def __call__(self, results, sampling_num, pre_selected=None, ref_results=None):
        mix_sample_list = []
        mix_selected_list = []

        if pre_selected is None:
            pre_selected = list(range(len(results)))    

        for sampling_fn in self.sampling_fns:
            sample_list, selected_list = sampling_fn(results, sampling_num, pre_selected, ref_results)

            mix_sample_list.extend(sample_list)
            mix_selected_list.extend(selected_list)

            # 更新pre_selected
            # 从pre_selected 中抛出selected_list
            pre_selected = list(set(pre_selected)-set(selected_list))

        return mix_sample_list, mix_selected_list


@UNCERTAINTY_SAMPLING.register_module()
class SamplingByComposer(object):
    def __init__(self, sampling_cfgs) -> None:
        super().__init__()
        self.sampling_fns= []
        for cfg in sampling_cfgs:
            sampling_fn = build_from_cfg(cfg, UNCERTAINTY_SAMPLING)
            self.sampling_fns.append(sampling_fn)

    def __call__(self, results, sampling_num, pre_selected=None, ref_results=None):
        sample_list = []        
        if pre_selected is None:
            pre_selected = list(range(len(results)))    

        selected_list = pre_selected
        for sampling_fn in self.sampling_fns:
            sample_list, selected_list = sampling_fn(results, sampling_num, selected_list, ref_results)

        return sample_list, selected_list


class Activelearning(Tester):
    def __init__(self, cfg, work_dir='./', gpu_id=-1, distributed=False):
        self.cfg = cfg
        # 默认使用单卡进行主动学习即可
        assert(not distributed)

        # 远程下载历史数据        
        if getattr(cfg, 'root', '') != '':
            file_client_mkdir(self.cfg.root, True)
            dir_name = os.path.dirname(cfg.root)
            if file_client_exists(os.path.join(dir_name, 'activelearning')):
                file_client_get(os.path.join(dir_name, 'activelearning'), './')

        if not os.path.exists('./activelearning'):
            os.mkdir('./activelearning')

        # has finish round info
        has_finish_anns = []
        for sub_folder in os.listdir('./activelearning'):
            if os.path.exists(f'./activelearning/{sub_folder}/annotation.json'):
                has_finish_anns.append(f'./activelearning/{sub_folder}/annotation.json')

        unlabel_filter_warp = dict(
            type='IterableDatasetFilter',
            dataset=build_from_cfg(cfg.data.test, DATASETS, None),
            not_in_anns=has_finish_anns
        )

        cfg.data.test = unlabel_filter_warp
        super().__init__(cfg, work_dir, gpu_id, distributed)

        # 采样函数
        self.sampling_fn = build_from_cfg(cfg.model.init_cfg.sampling_fn, UNCERTAINTY_SAMPLING)
        self.sampling_count = cfg.model.init_cfg.sampling_count # 计算预测次数
        self.sampling_num = cfg.model.init_cfg.sampling_num     # 最终采样数（最有价值的等待标注样本）

    def select(self, exp_name):
        # 添加运行标记
        running_flag(self.cfg.get('root', None))

        rank, _ = get_dist_info()        
        sample_results = None
        for _ in range(self.sampling_count):
            if not self.distributed:
                results = single_gpu_test(self.model, self.data_loader, needed_info=['image_meta'])
            else:
                results = multi_gpu_test(self.model, self.data_loader, needed_info=['image_meta'])

            if rank == 0:
                # 仅在master节点，合并所有结果
                if sample_results is None:
                    sample_results = [{} for i in range(len(results))]
                    for i in range(len(results)):
                        for k,v in results[i].items():
                            sample_results[i][k] = [v]
                else:
                    for i in range(len(results)):
                        for k,v in results[i].items():
                            sample_results[i][k].append(v) 
                            
            # 同步
            if self.distributed:
                dist.barrier()
        
        if rank == 0:
            # 仅在master节点运行分析样本
            # step1: 根据当前模型，挑选出需要标注的样本
            sampling_list, _ = self.sampling_fn(sample_results, self.sampling_num, ref_results=None)

            # step2: 遍历数据找到挑选出的样本，并保存到目标            
            # step2.1: 清空处理管线
            from antgo.framework.helper.dataset import PIPELINES
            self.dataset.dataset.pipeline = [build_from_cfg(dict(type='Meta', keys=['image_file', 'tag']), PIPELINES)]  # 仅添加Meta处理管线
            self.dataset.dataset._fields = ["image", "image_meta"] #
            self.dataset.dataset._alias = ["image", "image_meta"] #

            # step2.2: 保存到目标目录
            target_folder = f'./activelearning/{exp_name}'
            if not os.path.exists(target_folder):
                os.makedirs(target_folder)

            image_folder = os.path.join(target_folder, 'images')
            if not os.path.exists(image_folder):
                os.makedirs(image_folder)

            # TODO，每个样本如何进行唯一标识
            selected_sample_ids = [f"{sample['tag']}/{sample['image_file']}" for sample in sampling_list]

            image_file_name_list = []
            compare_pos = 0
            for sample in self.dataset:
                sample_tag = sample['image_meta']['tag']
                sample_id = f"{sample_tag}/{sample['image_meta']['image_file']}"
                
                if sample_id == selected_sample_ids[compare_pos]:
                    compare_pos += 1
                    # 防止在原目录下image_file是多层级的，在此将/替换为-
                    file_name = f"{sample_tag}_{sample['image_meta']['image_file'].replace('/','-')}"
                    if not (file_name.lower().endswith('.png') or file_name.lower().endswith('.jpg') or file_name.lower().endswith('.jpeg')):
                        file_name = f'{file_name}.png'
                    image_file_name_list.append((file_name, sample_tag))
                    cv2.imwrite(os.path.join(image_folder, file_name), sample['image'])

            # step2.3: 生成解析文件
            # 生成默认格式解析json
            annotation_file_name = 'annotation.json'
            annotation_list = []
            # 加载默认标准格式
            sgt = SampleGTTemplate() 
            for sample_file_name, sample_file_tag in image_file_name_list:
                sample_file_path = os.path.join(image_folder, sample_file_name)
                width, height = imagesize.get(sample_file_path)

                sample_gt = sgt.get()
                sample_gt_cp = copy.deepcopy(sample_gt)
                sample_gt_cp.update({
                    'image_file': f'images/{sample_file_name}'
                })
                sample_gt_cp['tag'] = f'{sample_file_tag}'
                sample_gt_cp['height'] = height
                sample_gt_cp['width'] = width
                annotation_list.append(sample_gt_cp)

            with open(os.path.join(target_folder, annotation_file_name), 'w', encoding="utf-8") as fp:
                json.dump(annotation_list, fp)

            with open(os.path.join(target_folder, 'meta.json'), 'w', encoding="utf-8") as fp:
                meta_info= sgt.meta()
                meta_info['extent']['total'] = len(annotation_list)
                json.dump(meta_info, fp)

            # step2.4: 保存到远程
            if getattr(self.cfg, 'root', '') != '':
                # 记录本次挑选出的数据记录
                dir_name = os.path.dirname(self.cfg.root)
                target_path = os.path.join(dir_name, 'activelearning', exp_name)
                if not file_client_exists(target_path):
                    file_client_mkdir(target_path, True)
                file_client_put(target_path, os.path.join(target_folder, annotation_file_name), False)

                # 打包本次挑选出的数据，并保存到标注目录下
                # root/label/exp_name/data.tar
                target_path = os.path.join(dir_name, 'label', exp_name)
                if not file_client_exists(target_path):
                    file_client_mkdir(target_path, True)

                os.system(f'cd {target_folder} && tar -cf ../data.tar .')
                file_client_put(target_path, './activelearning/data.tar', False)
                os.system('rm ./activelearning/data.tar')

            # 添加完成标记
            finish_flag(self.cfg.get('root', None))


import torch
import torch.nn as nn
from antgo.framework.helper.cnn import bias_init_with_prob, normal_init
from antgo.framework.helper.models.builder import HEADS, build_loss
from antgo.framework.helper.models.detectors.core.utils import multi_apply
from torchvision.ops import batched_nms 
from ..utils.gaussian_target import gaussian_radius, gen_gaussian_target
from ..utils.gaussian_target import (get_local_maximum, get_topk_from_heatmap,
                                     transpose_and_gather_feat)
from .base_dense_head import BaseDenseHead
import torch.nn.functional as F


def iou_loss(preds, targets):
    '''
    Args:
    preds: [n,4] ltrb
    targets: [n,4]
    '''
    lt = torch.min(preds[:, :2], targets[:, :2])
    rb = torch.min(preds[:, 2:], targets[:, 2:])
    wh = (rb + lt).clamp(min=0)
    overlap = wh[:, 0] * wh[:, 1]  # [n]
    area1 = (preds[:, 2] + preds[:, 0]) * (preds[:, 3] + preds[:, 1])
    area2 = (targets[:, 2] + targets[:, 0]) * (targets[:, 3] + targets[:, 1])
    iou = overlap / (area1 + area2 - overlap)
    loss = -iou.clamp(min=1e-6).log()
    return loss.sum()

def giou_loss(preds, targets):
    '''
    Args:
    preds: [n,4] ltrb
    targets: [n,4]
    '''
    lt_min = torch.min(preds[:, :2], targets[:, :2])
    rb_min = torch.min(preds[:, 2:], targets[:, 2:])
    wh_min = (rb_min + lt_min).clamp(min=0)
    overlap = wh_min[:, 0] * wh_min[:, 1]  # [n]
    area1 = (preds[:, 2] + preds[:, 0]) * (preds[:, 3] + preds[:, 1])
    area2 = (targets[:, 2] + targets[:, 0]) * (targets[:, 3] + targets[:, 1])
    union = (area1 + area2 - overlap)
    iou = overlap / union

    lt_max = torch.max(preds[:, :2], targets[:, :2])
    rb_max = torch.max(preds[:, 2:], targets[:, 2:])
    wh_max = (rb_max + lt_max).clamp(0)
    G_area = wh_max[:, 0] * wh_max[:, 1]  # [n]

    giou = iou - (G_area - union) / G_area.clamp(1e-10)
    loss = 1. - giou
    return loss.sum()


@HEADS.register_module()
class FcosHeadML(BaseDenseHead):
    """Objects as Points Head. CenterHead use center_point to indicate object's
    position. Paper link <https://arxiv.org/abs/1904.07850>

    Args:
        in_channel (int): Number of channel in the input feature map.
        feat_channel (int): Number of channel in the intermediate feature map.
        num_classes (int): Number of categories excluding the background
            category.
        loss_center_heatmap (dict | None): Config of center heatmap loss.
            Default: GaussianFocalLoss.
        loss_wh (dict | None): Config of wh loss. Default: L1Loss.
        loss_offset (dict | None): Config of offset loss. Default: L1Loss.
        train_cfg (dict | None): Training config. Useless in CenterNet,
            but we keep this variable for SingleStageDetector. Default: None.
        test_cfg (dict | None): Testing config of CenterNet. Default: None.
        init_cfg (dict or list[dict], optional): Initialization config dict.
            Default: None
    """

    def __init__(
        self,
        in_channel,
        feat_channel,
        num_classes,
        down_stride=[8,16],
        rescale=1.0,
        score_thresh=0.15,
        loss_ch=dict(
            type='GaussianFocalLoss', loss_weight=1.0),
        loss_rg=dict(
            type='IouLoss', loss_weight=0.2),
        train_cfg=dict(
            limit_range={8:[-1, 64], 16: [64,128], 32: [128,256], 64: [256,512],128: [512, 999999]}),
        test_cfg=None,
        init_cfg=dict(
            type='Xavier', layer='Conv2d', distribution='uniform')):
        super(FcosHeadML, self).__init__(init_cfg)
        self.num_classes = num_classes
        self.rescale_x, self.rescale_y = rescale if type(rescale) == list or type(rescale) == tuple else (rescale, rescale)
        self.score_thresh = score_thresh
        self.down_stride = down_stride
        self.limit_range = train_cfg.get('limit_range')

        self.in_channel = in_channel  
        self.feat_channel = feat_channel
        self._build_head()
        self.loss_center_heatmap=build_loss(loss_ch)
        
        self.loss_reg_func = iou_loss if loss_rg.get('type') == 'IouLoss' else giou_loss
        self.loss_reg_weight = loss_rg.get('loss_weight')
        self.train_cfg = train_cfg
        self.test_cfg = test_cfg      

    def _build_head(self):
        self.heatmap_head_list = nn.ModuleList()
        self.reg_head_list = nn.ModuleList()
        for _ in self.down_stride:
            self.heatmap_head_list.append(
                nn.Sequential(
                    nn.Conv2d(self.in_channel, self.feat_channel, kernel_size=3, stride=1, padding=1, bias=False),
                    nn.BatchNorm2d(self.feat_channel),
                    nn.ReLU(inplace=True),
                    nn.Conv2d(self.feat_channel, self.num_classes, kernel_size=1, bias=True)
                )
            )

            self.reg_head_list.append(
                nn.Sequential(
                    nn.Conv2d(self.in_channel,self.feat_channel,kernel_size=3,padding=1,bias=False),
                    nn.BatchNorm2d(self.feat_channel),
                    nn.ReLU(inplace=True),
                    nn.Conv2d(self.feat_channel, 4, kernel_size=3, padding=1, bias=False),
                    nn.BatchNorm2d(4),
                    nn.ReLU(inplace=True)
                )
            )

    def init_weights(self):
        """Initialize weights of the head."""
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(
                    m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0.0001)
                nn.init.constant_(m.running_mean, 0)
            elif isinstance(m, nn.BatchNorm1d):
                nn.init.constant_(m.weight, 1)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0.0001)
                nn.init.constant_(m.running_mean, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def forward(self, feats):
        """Forward features. Notice CenterNet head does not use FPN.

        Args:
            feats (tuple[Tensor]): Features from the upstream network, each is
                a 4D-tensor.

        Returns:
            center_heatmap_preds (List[Tensor]): center predict heatmaps for
                all levels, the channels number is num_classes.
            wh_preds (List[Tensor]): wh predicts for all levels, the channels
                number is 2.
            offset_preds (List[Tensor]): offset predicts for all levels, the
               channels number is 2.
        """
        center_heatmap_preds = []
        reg_preds = []
        # for feat, heatmap_head_func, reg_head_func in zip(feats, self.heatmap_head_list, self.reg_head_list):
        #     center_heatmap_preds.append(heatmap_head_func(feat))
        #     reg_preds.append(reg_head_func(feat))
        for level_i in range(len(feats)):
            center_heatmap_preds.append(self.heatmap_head_list[level_i](feats[level_i]))
            reg_preds.append(self.reg_head_list[level_i](feats[level_i]))

        return center_heatmap_preds, reg_preds

    def loss(self,
             center_heatmap_preds,
             reg_preds,
             bboxes,
             labels,
             image_meta, gt_bboxes_ignore=None):

        loss_reg_offset_avg = []
        loss_center_heatmap_avg = []

        pred_reg_offset_multi_levels = []
        gt_reg_offset_multi_levels = []
        mask_multi_levels = []

        batch_size = center_heatmap_preds[0].shape[0]

        for level_i, (center_heatmap_pred, reg_pred) in enumerate(zip(center_heatmap_preds, reg_preds)):
            target_result, avg_factor = \
                self.get_targets(bboxes, labels, center_heatmap_pred.shape, image_meta, self.down_stride[level_i], center_heatmap_pred.device)
            
            center_heatmap_target = target_result['center_heatmap_target']
            reg_targets = target_result['reg_targets']
            reg_weights = target_result['reg_weights']

            loss_center_heatmap = \
                self.loss_center_heatmap(center_heatmap_pred.sigmoid(), center_heatmap_target, avg_factor=avg_factor)
            loss_center_heatmap_avg.append(loss_center_heatmap.view(1))
            
            pred = reg_pred.permute(0, 2, 3, 1)
            pred = torch.reshape(pred, [batch_size, -1, 4]) * self.down_stride[level_i]
            pred_reg_offset_multi_levels.append(pred)

            gt = reg_targets.permute(0, 2, 3, 1)
            gt = torch.reshape(gt, [batch_size, -1, 4])
            gt_reg_offset_multi_levels.append(gt)
            mask_multi_levels.append(reg_weights.view(batch_size, -1))

        loss_center_heatmap_avg = torch.mean(torch.cat(loss_center_heatmap_avg))

        loss_reg_offset_avg = []
        pred_reg_offset_multi_levels = torch.cat(pred_reg_offset_multi_levels, dim=1)
        gt_reg_offset_multi_levels = torch.cat(gt_reg_offset_multi_levels, dim=1)
        mask_multi_levels = torch.cat(mask_multi_levels, dim=1)

        for batch_index in range(batch_size):
            pred_pos = pred_reg_offset_multi_levels[batch_index][mask_multi_levels[batch_index].to(torch.bool)]#[num_pos_b,4]
            target_pos=gt_reg_offset_multi_levels[batch_index][mask_multi_levels[batch_index].to(torch.bool)]#[num_pos_b,4]
            loss_reg_offset_avg.append(self.loss_reg_func(pred_pos,target_pos).view(1))

        loss_reg_offset_avg = torch.mean(torch.cat(loss_reg_offset_avg))
        total_loss = dict(
            loss_center_heatmap=loss_center_heatmap_avg,
            loss_reg=loss_reg_offset_avg * self.loss_reg_weight)

        return total_loss

    def coords_fmap(self, h, w, stride):
        shifts_x = torch.arange(0, w * stride, stride, dtype=torch.float32)
        shifts_y = torch.arange(0, h * stride, stride, dtype=torch.float32)

        shift_y, shift_x = torch.meshgrid(shifts_y, shifts_x)
        shift_x = torch.reshape(shift_x, [-1])
        shift_y = torch.reshape(shift_y, [-1])
        coords = torch.stack([shift_x, shift_y], -1) + stride // 2
        return coords

    def get_targets(self, gt_bboxes, gt_labels, feat_shape, image_meta, stride, device):
        """Compute regression and classification targets in multiple images.

        Args:
            gt_bboxes (list[Tensor]): Ground truth bboxes for each image with
                shape (num_gts, 4) in [tl_x, tl_y, br_x, br_y] format.
            gt_labels (list[Tensor]): class indices corresponding to each box.
            img_shape (list[int]): image shape in [h, w] format.

        Returns:
            tuple[dict,float]: The float value is mean avg_factor, the dict has
               components below:
               - center_heatmap_target (Tensor): targets of center heatmap, \
                   shape (B, num_classes, H, W).
               - wh_target (Tensor): targets of wh predict, shape \
                   (B, 2, H, W).
               - offset_target (Tensor): targets of offset predict, shape \
                   (B, 2, H, W).
               - wh_offset_target_weight (Tensor): weights of wh and offset \
                   predict, shape (B, 2, H, W).
        """

        bs, _, feat_h, feat_w = feat_shape

        # img_h = image_meta[0]['image_shape'][0]
        # img_w = image_meta[0]['image_shape'][1]
        # width_ratio = float(feat_w / img_w)
        # height_ratio = float(feat_h / img_h)

        center_heatmap_target = torch.zeros([bs, self.num_classes, feat_h, feat_w], dtype=torch.float32, device=device)
        reg_target_list = []
        reg_weight_list = []
        for batch_id in range(bs):
            gt_bbox = gt_bboxes[batch_id]       # gt_bbox shape: Nx4
            num_gt_bbox = gt_bbox.shape[0]
            if num_gt_bbox == 0:
                # 防止无目标样本
                reg_target_list.append(torch.zeros([4, feat_h, feat_w], dtype=torch.float32, device=device))
                reg_weight_list.append(torch.zeros([1, feat_h, feat_w], dtype=torch.float32, device=device))
                continue

            coords = self.coords_fmap(h=feat_h, w=feat_w, stride=stride).to(device=center_heatmap_target.device) #[h*w,2]
            x = coords[:, 0]
            y = coords[:, 1]
            l_off = x[:,None] - gt_bbox[...,0][None,:]    #[h*w,1]-[1,m]-->[h*w,m]
            t_off = y[:,None] - gt_bbox[...,1][None,:]
            r_off = gt_bbox[...,2][None,:] - x[:,None]
            b_off = gt_bbox[...,3][None,:] - y[:,None]
            ltrb_off = torch.stack([l_off,t_off,r_off,b_off],dim=-1)                        #[h*w,m,4]
            areas = (ltrb_off[...,0]+ltrb_off[...,2])*(ltrb_off[...,1]+ltrb_off[...,3])     #[h*w,m]

            off_min = torch.min(ltrb_off,dim=-1)[0]     #[h*w,m]
            off_max = torch.max(ltrb_off,dim=-1)[0]     #[h*w,m]
            mask_in_gtboxes = off_min>0
            mask_in_level = \
                (off_max>self.limit_range[stride][0])&(off_max<=self.limit_range[stride][1])   # 设置最大尺寸和最小尺寸约束

            radiu = stride * 1.5
            gt_center_x = (gt_bbox[...,0]+gt_bbox[...,2])/2
            gt_center_y = (gt_bbox[...,1]+gt_bbox[...,3])/2
            c_l_off = x[:,None] - gt_center_x[None,:]#[h*w,1]-[1,m]-->[h*w,m]
            c_t_off = y[:,None] - gt_center_y[None,:]
            c_r_off = gt_center_x[None,:] - x[:,None]
            c_b_off = gt_center_y[None,:] - y[:,None]
            c_ltrb_off = torch.stack([c_l_off,c_t_off,c_r_off,c_b_off],dim=-1)#[h*w,m,4]
            c_off_max = torch.max(c_ltrb_off,dim=-1)[0]
            mask_center = c_off_max<radiu

            mask_pos=mask_in_gtboxes&mask_in_level&mask_center      #[h*w,m]
            areas[~mask_pos] = 99999999
            areas_min_ind=torch.min(areas,dim=-1)[1]    #[h*w]

            reg_target = ltrb_off[
                torch.zeros_like(areas,dtype=torch.bool).scatter_(-1, areas_min_ind.unsqueeze(dim=-1),1)]   #[h*w,4]
            reg_target = torch.reshape(reg_target,(-1,4))       #[h*w,4]
            mask_pos_2 = mask_pos.long().sum(dim=-1)            #[h*w]
            mask_pos_2 = mask_pos_2 >= 1

            gt_label = gt_labels[batch_id]  # m

            # 测试不使用高斯
            gt_label = torch.broadcast_tensors(gt_label[None, :], areas.long())[0]  # [h*w,m]            
            cls_targets = gt_label[
                torch.zeros_like(areas,dtype=torch.bool).scatter_(-1, areas_min_ind.unsqueeze(dim=-1), 1)
            ]
            cls_targets = F.one_hot(cls_targets, num_classes=self.num_classes)
            cls_targets[~(mask_pos_2.to(torch.bool))] = 0            
            cls_targets = torch.permute(cls_targets, (1,0))
            cls_targets = torch.reshape(cls_targets, (-1, feat_h, feat_w))
            center_heatmap_target[batch_id] = cls_targets

            # reshape
            reg_target[~(mask_pos_2.to(torch.bool))] = 0            
            reg_target = reg_target.reshape(feat_h, feat_w, 4).permute(2,0,1)
            mask_pos_2 = mask_pos_2.to(torch.float32).reshape(1, feat_h,feat_w)

            # # 基于高斯构建目标
            # center_x = (gt_bbox[:, [0]] + gt_bbox[:, [2]]) * width_ratio / 2
            # center_y = (gt_bbox[:, [1]] + gt_bbox[:, [3]]) * height_ratio / 2
            # gt_centers = torch.cat((center_x, center_y), dim=1)            
            # for j, ct in enumerate(gt_centers):
            #     ctx_int, cty_int = ct.int()
            #     if mask_pos[cty_int*feat_w+ctx_int, j] == 0:
            #         continue
            #     scale_box_h = (gt_bbox[j][3] - gt_bbox[j][1]) * height_ratio
            #     scale_box_w = (gt_bbox[j][2] - gt_bbox[j][0]) * width_ratio

            #     radius = gaussian_radius([scale_box_h, scale_box_w], min_overlap=0.6)
            #     radius = max(int(2), int(radius))

            #     ind = gt_label[j]
            #     gen_gaussian_target(center_heatmap_target[batch_id, ind], [ctx_int, cty_int], radius)

            reg_target_list.append(reg_target)
            reg_weight_list.append(mask_pos_2)

        reg_targets = torch.stack(reg_target_list, 0)           #Bx4xHxW
        reg_weights = torch.stack(reg_weight_list, 0)

        avg_factor = max(1, center_heatmap_target.eq(1).sum())
        target_result = dict(
            center_heatmap_target=center_heatmap_target,
            reg_targets=reg_targets,
            reg_weights=reg_weights)
        return target_result, avg_factor

    def get_bboxes(self,
                   center_heatmap_preds,
                   reg_preds,
                   image_meta,
                   rescale=True,
                   with_nms=False):
        """Transform network output for a batch into bbox predictions.

        Args:
            center_heatmap_preds (list[Tensor]): Center predict heatmaps for
                all levels with shape (B, num_classes, H, W).
            wh_preds (list[Tensor]): WH predicts for all levels with
                shape (B, 2, H, W).
            offset_preds (list[Tensor]): Offset predicts for all levels
                with shape (B, 2, H, W).
            img_metas (list[dict]): Meta information of each image, e.g.,
                image size, scaling factor, etc.
            rescale (bool): If True, return boxes in original image space.
                Default: True.
            with_nms (bool): If True, do nms before return boxes.
                Default: False.

        Returns:
            list[tuple[Tensor, Tensor]]: Each item in result_list is 2-tuple.
                The first item is an (n, 5) tensor, where 5 represent
                (tl_x, tl_y, br_x, br_y, score) and the score between 0 and 1.
                The shape of the second tensor in the tuple is (n,), and
                each element represents the class label of the corresponding
                box.
        """
        # multi level
        assert len(center_heatmap_preds) == len(reg_preds)
        result_list = []
        for img_id in range(len(image_meta)):
            level_det_bboxes = []
            level_det_labels = []
            for level_i, (center_heatmap_pred, reg_pred) in enumerate(zip(center_heatmap_preds, reg_preds)):
                det_bboxes, det_labels = self._get_bboxes_single(
                    center_heatmap_pred[img_id:img_id + 1, ...],
                    reg_pred[img_id:img_id + 1, ...],
                    image_meta[img_id],
                    stride=self.down_stride[level_i])
                
                level_det_bboxes.append(det_bboxes)
                level_det_labels.append(det_labels)

            det_bboxes = torch.concat(level_det_bboxes, 0)
            det_labels = torch.concat(level_det_labels, 0)
            if rescale:
                det_bboxes[..., :4] *= det_bboxes.new_tensor((self.rescale_x, self.rescale_y, self.rescale_x, self.rescale_y))

            if 'nms' in self.test_cfg or with_nms:
                det_bboxes, det_labels = \
                    self._bboxes_nms(det_bboxes, det_labels, self.test_cfg)

            result_list.append((det_bboxes, det_labels))
        return result_list

    def _get_bboxes_single(self,
                           center_heatmap_pred,
                           reg_pred,
                           img_meta, stride):
        """Transform outputs of a single image into bbox results.

        Args:
            center_heatmap_pred (Tensor): Center heatmap for current level with
                shape (1, num_classes, H, W).
            wh_pred (Tensor): WH heatmap for current level with shape
                (1, num_classes, H, W).
            offset_pred (Tensor): Offset for current level with shape
                (1, corner_offset_channels, H, W).
            img_meta (dict): Meta information of current image, e.g.,
                image size, scaling factor, etc.
            rescale (bool): If True, return boxes in original image space.
                Default: False.
            with_nms (bool): If True, do nms before return boxes.
                Default: True.

        Returns:
            tuple[Tensor, Tensor]: The first item is an (n, 5) tensor, where
                5 represent (tl_x, tl_y, br_x, br_y, score) and the score
                between 0 and 1. The shape of the second tensor in the tuple
                is (n,), and each element represents the class label of the
                corresponding box.
        """
        center_heatmap_pred = torch.sigmoid(center_heatmap_pred)
        batch_det_bboxes, batch_labels = self.decode_heatmap(
            center_heatmap_pred,
            reg_pred,
            img_meta['image_shape'],            # batch_input_shape->input_shape
            k=self.test_cfg.topk,
            kernel=self.test_cfg.local_maximum_kernel, 
            stride=stride)

        det_bboxes = batch_det_bboxes.view([-1, 5])
        det_labels = batch_labels.view(-1)

        # only for debug
        # image = cv2.imread(img_meta['image_file'])
        # det_bboxes_numpy = det_bboxes.detach().cpu().numpy()
        # for i in range(det_bboxes_numpy.shape[0]):
        #     x0,y0,x1,y1,_ = det_bboxes_numpy[i]
        #     cv2.rectangle(image, ((int)(x0),(int)(y0)), ((int)(x1),(int)(y1)), (255,0,0), 2)
        # cv2.imwrite('./aa.png', image)
        return det_bboxes, det_labels

    def decode_heatmap(self,
                       center_heatmap_pred,
                       reg_pred,
                       img_shape,
                       k=100,
                       kernel=3, stride=8):
        """Transform outputs into detections raw bbox prediction.

        Args:
            center_heatmap_pred (Tensor): center predict heatmap,
               shape (B, num_classes, H, W).
            wh_pred (Tensor): wh predict, shape (B, 2, H, W).
            offset_pred (Tensor): offset predict, shape (B, 2, H, W).
            img_shape (list[int]): image shape in [h, w] format.
            k (int): Get top k center keypoints from heatmap. Default 100.
            kernel (int): Max pooling kernel for extract local maximum pixels.
               Default 3.

        Returns:
            tuple[torch.Tensor]: Decoded output of CenterNetHead, containing
               the following Tensors:

              - batch_bboxes (Tensor): Coords of each box with shape (B, k, 5)
              - batch_topk_labels (Tensor): Categories of each box with \
                  shape (B, k)
        """
        center_heatmap_pred = get_local_maximum(center_heatmap_pred, kernel=kernel)

        *batch_dets, topk_ys, topk_xs = get_topk_from_heatmap(
            center_heatmap_pred, k=k)
        batch_scores, batch_index, batch_topk_labels = batch_dets

        ltrb_off = transpose_and_gather_feat(reg_pred, batch_index)     # BxHxWx4
        tl_x = (topk_xs * stride +  stride // 2 - ltrb_off[..., 0] * stride)
        tl_y = (topk_ys * stride +  stride // 2 - ltrb_off[..., 1] * stride)
        br_x = (topk_xs * stride +  stride // 2 + ltrb_off[..., 2] * stride)
        br_y = (topk_ys * stride +  stride // 2 + ltrb_off[..., 3] * stride)

        batch_bboxes = torch.stack([tl_x, tl_y, br_x, br_y], dim=2)
        batch_bboxes = torch.cat((batch_bboxes, batch_scores[..., None]),
                                 dim=-1)
        
        return batch_bboxes, batch_topk_labels

    def _bboxes_nms(self, bboxes, labels, cfg):
        if labels.numel() > 0:
            max_num = cfg.max_per_img
            keep = batched_nms(
                bboxes[:, :4], 
                bboxes[:, -1].contiguous(),
                labels, 
                cfg.nms)
            if max_num > 0:
                bboxes = bboxes[keep][:max_num]
                labels = labels[keep][:max_num]

            remained_index = []
            for i in range(bboxes.shape[0]):
                if bboxes[i,4] > self.score_thresh:
                    remained_index.append(i)
            
            bboxes = bboxes[remained_index]
            labels = labels[remained_index]
        return bboxes, labels



# -*- coding: UTF-8 -*-
# @Time    : 2021/11/13 6:13 下午
# @File    : ensemble_server.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from copyreg import pickle
import tornado.httpserver
import tornado.ioloop
import tornado.options
from tornado.options import define, options
from tornado import web, gen
from tornado import httpclient
import tornado.web
import os
import shutil
import signal
from antgo.crowdsource.base_server import *
from antgo.utils import logger
import sys
import uuid
import json
import time
import math
import numpy as np
from io import BytesIO
from tornado.concurrent import run_on_executor
from concurrent.futures import ThreadPoolExecutor
import threading
import pickle
import copy


def softmax(x, axis=None):
  x = x - x.max(axis=axis, keepdims=True)
  y = np.exp(x)
  return y / y.sum(axis=axis, keepdims=True)


MB = 1024 * 1024
GB = 1024 * MB
MAX_STREAMED_SIZE = 1*GB
@tornado.web.stream_request_body
class AverageApiHandler(BaseHandler):
  executor = ThreadPoolExecutor(16)

  def __init__(self, *args, **kwargs):
    super(AverageApiHandler, self).__init__(*args, **kwargs)
    self.save_seek = 0
    self.worker_prefix = None
    self.file_folder = None
    self.file_name = None

  @property
  def static_folder(self):
    return self.settings.get('static_path')

  def prepare(self):        
    # 设置最大流式数据量
    self.request.connection.set_max_body_size(MAX_STREAMED_SIZE)    
    headers = self.request.headers
    if 'worker_prefix' not in headers:
      return
        
    self.worker_prefix = headers['worker_prefix']   # 用于标识每个模型
    self.file_folder = os.path.join(self.static_folder, f"{self.worker_prefix}")
    if not os.path.exists(self.file_folder):
      os.makedirs(self.file_folder)
    self.file_name = f"{headers['id']}"             # variable ensemble.id

  def data_received(self, chunk):
      if self.file_name:
        if not os.path.exists(os.path.join(self.file_folder, self.file_name)):
          with open(os.path.join(self.file_folder, self.file_name), 'wb') as up:
              pass

        with open(os.path.join(self.file_folder, self.file_name), 'rb+') as up:
          up.seek(self.save_seek)
          up.write(chunk)
          self.save_seek = self.save_seek + len(chunk)

  def cal_shannon_entropy(self, preds, axis):
    uncertainty = -1.0 * np.sum(preds * np.log(preds + 1e-6), axis=axis, keepdims=True)
    return uncertainty

  @run_on_executor
  def avg(self, sample_id, value, weight):
    # 1.step 如果value=None，则直接进入等待聚合结果过程
    if value is None:
      self.consume_cond.acquire()
      while True:
        # 检查sample_id数据是否已经完成融合
        if 'sample' not in self.db or \
          sample_id not in self.db['sample'] or \
          self.db['sample'][sample_id]['ensemble'] == '':
          self.consume_cond.wait()
        else:
          break
      
      self.consume_cond.release()
      with open(self.db['sample'][sample_id]['ensemble'], 'rb') as fp:
        return fp.read()
    
    ###########################################################################
    # 2.step 检查sample_id是否已经在聚合过程中，或者已经完成聚合
    self.lock.acquire()
    is_in_process = 'sample' in self.db and \
        sample_id in self.db['sample'] and \
      'worker' in self.db['sample'] and \
      self.request.headers['worker_prefix'] in self.db['sample']['worker']
    self.lock.release()
    
    if is_in_process:
      # 仅有第二次调用才会进入这里
      self.consume_cond.acquire()
      while True:
        # 检查sample_id数据是否已经完成融合
        if self.db['sample'][sample_id]['ensemble'] == '':
          self.consume_cond.wait()          
        else:
          break
      
      self.consume_cond.release()
      with open(self.db['sample'][sample_id]['ensemble'], 'rb') as fp:
        return fp.read()
    ###########################################################################
    
    # 3.step sample_id 首次出现，进入积累数据并等待聚合过程    
    self.lock.acquire()
    # 数据积累操作（线程安全）
    if 'sample' not in self.db:
      self.db['sample'] = {}

    
    if sample_id not in self.db['sample']:
      self.db['sample'][sample_id] = {
        'data': [], 
        'count': self.producer_num, 
        'weight': [], 
        'worker': self.request.headers['worker_prefix'],
        'ensemble': ''}

    self.db['sample'][sample_id]['data'].append(value)
    self.db['sample'][sample_id]['weight'].append(weight)
    self.lock.release()
    
    self.cond.acquire()
    while True:
      if len(self.db['sample'][sample_id]['data']) == self.producer_num:
        self.cond.notify()
        break

      if len(self.db['sample'][sample_id]['data']) != self.producer_num:
        self.cond.wait()

    self.cond.release()

    # 到此，数据已经完备，开始做融合处理
    data = {}
    if self.uncertain_vote_cfg is not None and \
        len(self.db['sample'][sample_id]['data']) > 1:
      # 计算不确定度，并加权平均
      uncertain_axis = self.uncertain_vote_cfg['axis']
      all_uncertainty_org = []
      all_uncertainty_region = []
      
      keys = self.db['sample'][sample_id]['data'][0].keys()
      for key in keys:
        for sample_data_map, sample_weight_map in \
            zip(self.db['sample'][sample_id]['data'], self.db['sample'][sample_id]['weight']):
          
          sample_data = sample_data_map[key]
          sample_weight = sample_weight_map[key]
          # uncertainty_org 1,1,H,W
          uncertainty_org = self.cal_shannon_entropy(sample_data, uncertain_axis)
          uncertainty_region = np.ones(uncertainty_org.shape, dtype=np.float32)
          if self.uncertain_vote_cfg['thres'] != -1:
            uncertainty_region = (uncertainty_org > self.uncertain_vote_cfg['thres']).astype(np.float32)

          all_uncertainty_org.append(uncertainty_org)
          all_uncertainty_region.append(uncertainty_region)

        all_uncertainty_org = np.concatenate(all_uncertainty_org, axis=uncertain_axis)
        all_uncertainty_org = softmax(all_uncertainty_org, uncertain_axis)
        # certainty prob
        all_certainty_org = \
          np.split(1.0 - all_uncertainty_org,
                  indices_or_sections=all_uncertainty_org.shape[uncertain_axis],
                  axis=uncertain_axis)

        reweight_sample_weight = []
        for sample_weight_map, sample_certainty, sample_limited_region in \
            zip(self.db['sample'][sample_id]['weight'], all_certainty_org, all_uncertainty_region):
          sample_weight = sample_weight_map[key]
          certainty = sample_certainty * sample_limited_region + (1.0-sample_limited_region)
          reweight_sample_weight.append(sample_weight * certainty)

        weighted_data = 0.0
        weighted_norm = 0.0
        for sample_data_map, sample_weight in zip(self.db['sample'][sample_id]['data'],
                                              reweight_sample_weight):
          sample_data = sample_data_map[key]
          weighted_data += sample_data * sample_weight
          weighted_norm += sample_weight

        data[key] = weighted_data / weighted_norm
    else:
      # 加权平均
      weighted_data = 0.0
      weighted_norm = 0.0
      keys = self.db['sample'][sample_id]['data'][0].keys()
      
      for key in keys:
        for sample_data_map, sample_weight in zip(self.db['sample'][sample_id]['data'],
                                              self.db['sample'][sample_id]['weight']):
          sample_data = sample_data_map[key]
          weighted_data += sample_data * sample_weight
          weighted_norm += sample_weight

        data[key] = weighted_data / weighted_norm

    # 清空缓存 (线程安全)
    self.lock.acquire()
    self.db['sample'][sample_id]['count'] -= 1
    assert(self.db['sample'][sample_id]['count'] >= 0)
    if self.db['sample'][sample_id]['count'] == 0:
      # 保存聚合后结果
      self.consume_cond.acquire()
      if not os.path.exists(os.path.join(self.static_folder, 'ensemble')):
        os.makedirs(os.path.join(self.static_folder, 'ensemble'))
            
      ensemble_data_file_path = os.path.join(self.static_folder, 'ensemble', f"{sample_id}")
      with open(ensemble_data_file_path, 'wb') as fp:
        pickle.dump(data, fp)  
      
      self.db['sample'][sample_id].update({'data': [], 'count': 0, 'ensemble': ensemble_data_file_path})
      # 已经完成聚合数据的保存，通知可能处于等待的线程
      self.consume_cond.notifyAll()
      self.consume_cond.release()

    # 删除临时数据
    if os.path.exists(os.path.join(self.file_folder, self.file_name)):
      os.remove(os.path.join(self.file_folder, self.file_name))
    self.lock.release()
    return pickle.dumps(data)

  @gen.coroutine
  def post(self, *args, **kwargs):
    # id,name,value
    id = self.request.headers.get('id')
    feedback = self.request.headers.get('feedback')

    value = None
    weight = 1.0
    if self.worker_prefix is not None:
      if not os.path.exists(os.path.join(self.file_folder, self.file_name)):
        self.response(RESPONSE_STATUS_CODE.REQUEST_INVALID)
        return

      # 模型数据
      with open(os.path.join(self.file_folder, self.file_name), 'rb') as fp:
        value = pickle.load(fp)
        
      # 模型权重
      weight = (float)(self.request.headers.get('weight', 1.0))

    # 计算加权值
    pickle_data_bytes = yield self.avg(id, value, weight)

    # 下载数据服务
    if feedback == 'True':
      self.set_header('Content-Type', 'application/octet-stream')
      self.set_header('Content-Disposition', f'attachment; filename=data')
      with BytesIO(pickle_data_bytes) as f:
        while True:
          data = f.read(MB)
          if not data:
              break
          self.write(data)

    self.finish()

  @property
  def cond(self):
    return self.settings['cond']

  @property
  def consume_cond(self):
    return self.settings['consume_cond']

  @property
  def producer_num(self):
      return self.settings['producer_num']

  @property
  def lock(self):
    return self.settings['lock']

  @property
  def uncertain_vote_cfg(self):
    return self.settings.get('uncertain_vote_cfg', None)


@tornado.web.stream_request_body
class PutGetApiHandler(BaseHandler):
  executor = ThreadPoolExecutor(16)
  cache = {}
  def __init__(self, *args, **kwargs):
    super(PutGetApiHandler, self).__init__(*args, **kwargs)
      
  def prepare(self):
    # 设置最大流式数据量
    self.request.connection.set_max_body_size(MAX_STREAMED_SIZE)
    headers = self.request.headers
    self.cond.acquire()
    if headers['id'] not in PutGetApiHandler.cache:
        PutGetApiHandler.cache[headers['id']] = {
            'status': False,
            'io':  BytesIO(),
            'consume': 0
        }
    self.cond.release()
          
  def data_received(self, chunk):            
    PutGetApiHandler.cache[self.request.headers['id']]['io'].write(chunk)
    
  @gen.coroutine
  def post(self):
    # here, 数据传输已经完成
    self.cond.acquire()
    PutGetApiHandler.cache[self.request.headers['id']]['status'] = True
    PutGetApiHandler.cache[self.request.headers['id']]['io'].seek(0, os.SEEK_SET)
    self.cond.notifyAll()
    self.cond.release()
    self.finish()
  
  @run_on_executor
  def wait_data_ready(self, id):
    self.cond.acquire()
    while True:
        if PutGetApiHandler.cache[id]['status']:
            break            
        self.cond.wait()
    self.cond.release()
    
    self.set_header('Content-Type', 'application/octet-stream')
    self.set_header('Content-Disposition', f'attachment; filename={id}')
    f = copy.deepcopy(PutGetApiHandler.cache[id]['io'])
    while True:
        data = f.read(MB)
        if not data:
            break
          
        self.write(data)
        
    PutGetApiHandler.cache[id]['consume'] += 1
    if PutGetApiHandler.cache[id]['consume'] == self.consumer_num:
        # 已经完成所有消费,删除数据
        PutGetApiHandler.cache.pop(id)
    
  @property
  def cond(self):
    return self.settings['put_get_cond']
  
  @gen.coroutine
  def get(self):
    id = self.request.headers['id']
    
    yield self.wait_data_ready(id)
    self.finish()
      
  @property
  def consumer_num(self):
    return self.settings['consumer_num']

@tornado.web.stream_request_body
class DataApiHandler(BaseHandler):
  @property
  def dump_dir(self):
    return self.settings.get('dump_dir', None)

  @gen.coroutine
  def get(self, file_name):
    # 提供变量数据服务
    #   ave_result = msgpack.packb(ave_result, default=ms.encode)
    file_name = file_name[1:]
    file_path = os.path.join(self.dump_dir, 'ensemble', 'merge', file_name)
    if not os.path.exists(file_path):
      self.set_status(404)
      return

    self.set_header('Content-Type', 'application/octet-stream')
    self.set_header('Content-Disposition', f'attachment; filename={file_name}')
    with open(file_path, 'rb') as f:
      while True:
        data = f.read(MB)
        if not data:
            break
        self.write(data)

    self.finish()


class LiveApiHandler(BaseHandler):
  @gen.coroutine
  def get(self):
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class GracefulExitException(Exception):
  @staticmethod
  def sigterm_handler(signum, frame):
    raise GracefulExitException()


def ensemble_server_start(
    dump_dir,
    server_port,
    worker_num,
    uncertain_vote_cfg=None):
  # uncertain_vote_cfg {'axis': 1, 'thres': 0}
  # register sig
  signal.signal(signal.SIGTERM, GracefulExitException.sigterm_handler)

  # 0.step define http server port
  define('port', default=server_port, help='run on port')

  try:
    # 静态资源目录
    static_dir = os.path.join(dump_dir, 'ensemble')
    if not os.path.exists(os.path.join(static_dir, 'static')):
      os.makedirs(os.path.join(static_dir, 'static'))

    # 2.step launch show server
    db = {}
    settings = {
      'static_path': os.path.join(static_dir, 'static'),
      'dump_dir': dump_dir,
      'port': server_port,
      'cookie_secret': str(uuid.uuid4()),
      'worker_num': worker_num,
      'cond': threading.Condition(),
      'consume_cond': threading.Condition(),
      'put_get_cond': threading.Condition(),
      'lock': threading.Lock(),
      'db': db,
      'uncertain_vote_cfg': uncertain_vote_cfg,
      'producer_num': worker_num,
      'consumer_num': worker_num
    }

    app = tornado.web.Application(handlers=[(r"/antgo/api/ensemble/avg/", AverageApiHandler),
                                            (r"/antgo/api/ensemble/put/", PutGetApiHandler),
                                            (r"/antgo/api/ensemble/get/", PutGetApiHandler),
                                            (r"/antgo/api/ping/", LiveApiHandler),
                                            (r"/antgo/api/ensemble/data/(.*)", DataApiHandler)],
                                  **settings)
    http_server = tornado.httpserver.HTTPServer(app)
    http_server.listen(options.port)

    logger.info('ensemble is providing server on port %d' % server_port)
    tornado.ioloop.IOLoop.instance().start()
    logger.info('ensemble stop server')
  except GracefulExitException:
    logger.info('ensemble server exit')
    sys.exit(0)
  except KeyboardInterrupt:
    pass


# Copyright (c) MLTALKER. All rights reserved.
import sys
sys.path.append('/root/paddlejob/workspace/env_run/portrait/antgo')
# sys.path.append('/root/paddlejob/workspace/env_run/portrait/InterHand26M/common')
# sys.path.append('/root/paddlejob/workspace/env_run/portrait/InterHand26M/main')

import argparse
import copy
import os
import os.path as osp
import time
import warnings
import numpy as np

import torch
import torch.distributed as dist
from antgo.framework.helper.utils import Config, DictAction
from antgo.framework.helper.runner import get_dist_info, init_dist
from antgo.framework.helper.utils import get_git_hash
from antgo.framework.helper.utils import (collect_env,update_data_root)
from antgo.framework.helper.utils import (collect_env, get_logger, update_data_root)
from antgo.framework.helper.utils.util_distribution import get_device
from antgo.framework.helper.runner.utils import *
from antgo.framework.helper.models.builder import *
from antgo.framework.helper.apis.train import *
from antgo.framework.helper.dataset import *
from antgo.framework.helper.tools.util import *


def main():
    args = parse_args()
    cfg = Config.fromfile(args.config)

    if args.auto_scale_lr:
        if 'auto_scale_lr' in cfg and \
                'enable' in cfg.auto_scale_lr and \
                'base_batch_size' in cfg.auto_scale_lr:
            cfg.auto_scale_lr.enable = True
        else:
            warnings.warn('Can not find "auto_scale_lr" or '
                          '"auto_scale_lr.enable" or '
                          '"auto_scale_lr.base_batch_size" in your'
                          ' configuration file. Please update all the '
                          'configuration files to mmdet >= 2.24.1.')

    # set cudnn_benchmark
    if cfg.get('cudnn_benchmark', False):
        torch.backends.cudnn.benchmark = True

    # work_dir is determined in this priority: CLI > segment in file > filename
    if args.work_dir is not None and args.work_dir != '':
        # update configs according to CLI args if args.work_dir is not None
        cfg.work_dir = args.work_dir
    elif cfg.get('work_dir', None) is None:
        # use config filename as default work_dir if cfg.work_dir is None
        cfg.work_dir = osp.join('./work_dirs',
                                osp.splitext(osp.basename(args.config))[0])

    if args.resume_from is not None:
        cfg.resume_from = args.resume_from
    cfg.auto_resume = args.auto_resume
    cfg.gpu_ids = [args.gpu_id]
    
    # init distributed env first, since logger depends on the dist info.
    if args.launcher == 'none':
        distributed = False
    else:
        distributed = True
        init_dist(**cfg.get('dist_params', {}))
        # re-set gpu_ids with distributed training mode
        _, world_size = get_dist_info()
        cfg.gpu_ids = range(world_size)

    # create work_dir
    if not os.path.exists(cfg.work_dir):
        os.makedirs(cfg.work_dir)

    # dump config
    cfg.dump(osp.join(cfg.work_dir, osp.basename(args.config)))
    # init the logger before other steps
    timestamp = time.strftime('%Y-%m-%dx%H-%M-%S', time.localtime())
    log_file = osp.join(cfg.work_dir, f'{timestamp}.log')
    logger = get_logger('model',log_file=log_file, log_level=cfg.log_level)

    # init the meta dict to record some important information such as
    # environment info and seed, which will be logged
    meta = dict()
    # log env info
    env_info_dict = collect_env()
    env_info = '\n'.join([(f'{k}: {v}') for k, v in env_info_dict.items()])
    dash_line = '-' * 60 + '\n'
    logger.info('Environment info:\n' + dash_line + env_info + '\n' +
                dash_line)
    meta['env_info'] = env_info
    meta['config'] = cfg.pretty_text
    # log some basic info
    logger.info(f'Distributed training: {distributed}')
    logger.info(f'Config:\n{cfg.pretty_text}')

    cfg.device = get_device()
    # set random seeds
    seed = init_random_seed(args.seed, device=cfg.device)
    seed = seed + dist.get_rank() if args.diff_seed else seed
    logger.info(f'Set random seed to {seed}, '
                f'deterministic: {args.deterministic}')
    set_random_seed(seed, deterministic=args.deterministic)
    cfg.seed = seed
    meta['seed'] = seed
    meta['exp_name'] = osp.basename(args.config)

    # 1.step 创建模型并初始化
    if args.ext_module != '':
        # 导入ext module
        load_extmodule(args.ext_module)

    model = \
        build_model(cfg.model, train_cfg=cfg.get('train_cfg', {}), test_cfg=cfg.get('test_cfg', {}))
    model.init_weights()

    # 2.step 创建数据集
    datasets = [build_dataset(cfg.data.train)]
    
    # 3.step 训练模型过程
    train_model(
        model,
        datasets,
        cfg,
        distributed=distributed,
        validate=(not args.no_validate),
        timestamp=timestamp,
        meta=meta)

if __name__ == '__main__':
    main()



# 优化器配置
optimizer = dict(type='SGD', lr=0.01,  weight_decay=5e-4, momentum=0.9, nesterov=True)
optimizer_config = dict(grad_clip=dict(max_norm=20))

# 学习率调度配置
lr_config = dict(
    policy='CosineAnnealing',
    min_lr=1e-5,
    warmup_by_epoch=False,
    warmup_iters=2000,
    warmup='linear'
)

# 日志配置
log_config = dict(
    interval=50, 
    hooks=[
        dict(type='TextLoggerHook'),
    ])

# 模型配置
model = dict(
    type='TTFNet',
    backbone=dict(
        type='ResNet',
        depth=50, 
        in_channels=3, 
        out_indices=[3,2,1]),      
    neck=dict(type="FPN", in_channels=[512, 1024, 2048], out_channels=256, num_outs=3, upsample_cfg=dict(mode='bilinear', scale_factor=2.0)),
    bbox_head=dict(
        type='FcosHeadML',
        in_channel=256,
        feat_channel=256,
        num_classes=80,
        down_stride=[8,16,32],
        score_thresh=0.05,
        train_cfg=None,
        test_cfg=dict(topk=100, local_maximum_kernel=3, nms=0.6, max_per_img=50),
        loss_ch=dict(type='GaussianFocalLoss', loss_weight=1.0)
    ),  
)

# checkpoint配置
checkpoint_config = dict(interval=1, out_dir='./output/')       

# 数据配置
data=dict(
    train=dict(
        type='COCO2017',
        train_or_test='train',
        dir='./coco_dataset',
        ext_params={'task_type': 'OBJECT-DETECTION'},
        pipeline=[
            dict(type="Rotation", degree=30),
            dict(type='ResizeByShort', short_size=800, max_size=1333),
            dict(type='ColorDistort', hue=[-5,5,0.5], saturation=[0.7,1.3,0.5], contrast=[0.7,1.3,0.5], brightness=[0.7,1.3,0.5]),
            dict(type='RandomFlipImage', swap_labels=[]),
            dict(type='INormalize', mean=[128.0,128.0,128.0], std=[128.0,128.0,128.0],to_rgb=False, keys=['image']),
            dict(type='Permute', to_bgr=False, channel_first=True)
        ],
        inputs_def=dict(
            fields = ["image", 'bboxes', 'labels', 'image_meta']
        )
    ),
    train_dataloader=dict(
        samples_per_gpu=8, 
        workers_per_gpu=2,
        drop_last=True,
        shuffle=True,
        ignore_stack=['image', 'bboxes', 'labels', 'image_meta']
    ),
    val=dict(
        type='COCO2017',
        train_or_test='val',
        ext_params={'task_type': 'OBJECT-DETECTION'},
        dir='./coco_dataset',
        pipeline=[
            dict(type='ResizeByShort', short_size=800, max_size=1333),
            dict(type='INormalize', mean=[128.0,128.0,128.0], std=[128.0,128.0,128.0],to_rgb=False, keys=['image']),
            dict(type='Permute', to_bgr=False, channel_first=True)
        ],
        inputs_def=dict(
            fields= ["image", 'bboxes', 'labels', 'image_meta']
        )
    ),
    val_dataloader=dict(
        samples_per_gpu=8,
        workers_per_gpu=2,
        drop_last=False,
        shuffle=False,
        ignore_stack=['image', 'bboxes', 'labels', 'image_meta']
    ),
    test=dict(
        type='COCO2017',
        train_or_test='val',
        ext_params={'task_type': 'OBJECT-DETECTION'},
        dir='./coco_dataset',
        pipeline=[
            dict(type='ResizeByShort', short_size=800, max_size=1333),
            dict(type='INormalize', mean=[128.0,128.0,128.0], std=[128.0,128.0,128.0],to_rgb=False, keys=['image']),
            dict(type='Permute', to_bgr=False, channel_first=True)
        ],
        inputs_def=dict(
            fields= ["image", 'bboxes', 'labels', 'image_meta']
        )
    ),
    test_dataloader=dict(
        samples_per_gpu=10,
        workers_per_gpu=1,
        drop_last=False,
        shuffle=False,
        ignore_stack=['image','bboxes', 'labels', 'image_meta']
    ),    
)

# 评估方案配置
evaluation=dict(
    out_dir='./output/', 
    interval=1, 
    metric=dict(
        type='COCOCompatibleEval', 
        categories=[{'name': f'{label}', 'id': label} for label in range(80)],
        without_background=False
    ), 
    save_best='AP@[ IoU=0.50:0.95 | area= all | maxDets=100 ]',
    rule='greater'
)

# 导出配置
export=dict(
    input_shape_list = [[1,3,800,800]],
    input_name_list=["image"],
    output_name_list=["heatmap", "offset"]
)

max_epochs = 60


# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from paddle.io import Dataset, IterableDataset
import paddle
import warnings
import bisect


class ConcatDataset(Dataset):
    r"""Dataset as a concatenation of multiple datasets.

    This class is useful to assemble different existing datasets.

    Arguments:
        datasets (sequence): List of datasets to be concatenated
    """

    @staticmethod
    def cumsum(sequence):
        r, s = [], 0
        for e in sequence:
            l = len(e)
            r.append(l + s)
            s += l
        return r

    def __init__(self, datasets):
        super(ConcatDataset, self).__init__()
        assert len(datasets) > 0, 'datasets should not be an empty iterable'
        self.datasets = list(datasets)
        for d in self.datasets:
            assert not isinstance(
                d, IterableDataset
            ), "ConcatDataset does not support IterableDataset"
        self.cumulative_sizes = self.cumsum(self.datasets)

    def __len__(self):
        return self.cumulative_sizes[-1]

    def __getitem__(self, idx):
        if idx < 0:
            if -idx > len(self):
                raise ValueError(
                    "absolute value of index should not exceed dataset length")
            idx = len(self) + idx
        dataset_idx = bisect.bisect_right(self.cumulative_sizes, idx)
        if dataset_idx == 0:
            sample_idx = idx
        else:
            sample_idx = idx - self.cumulative_sizes[dataset_idx - 1]
        return self.datasets[dataset_idx][sample_idx]

    @property
    def cummulative_sizes(self):
        warnings.warn(
            "cummulative_sizes attribute is renamed to "
            "cumulative_sizes",
            DeprecationWarning,
            stacklevel=2)
        return self.cumulative_sizes


def _accumulate(iterable, fn=lambda x, y: x + y):
    # _accumulate([1,2,3,4,5]) --> 1 3 6 10 15
    # _accumulate([1,2,3,4,5], operators.mul) --> 1 2 6 24 120
    it = iter(iterable)
    try:
        total = next(it)
    except StopIteration:
        return
    yield total
    for element in it:
        total = fn(total, element)
        yield total


class Subset(Dataset):
    r"""
    Subset of a dataset at specified indices.

    Arguments:
        dataset (Dataset): The whole Dataset
        indices (sequence): Indices in the whole set selected for subset
    """

    def __init__(self, dataset, indices):
        self.dataset = dataset
        self.indices = indices

    def __getitem__(self, idx):
        return self.dataset[self.indices[idx]]

    def __len__(self):
        return len(self.indices)


def random_split(dataset, lengths, generator=None):
    r"""
    Randomly split a dataset into non-overlapping new datasets of given lengths.
    Optionally fix the generator for reproducible results, e.g.:

    >>> random_split(range(10), [3, 7], generator=torch.Generator().manual_seed(42))

    Arguments:
        dataset (Dataset): Dataset to be split
        lengths (sequence): lengths of splits to be produced
        generator (Generator): from torch import default_generator, which is not use in paddle.
    """
    if sum(lengths) != len(dataset):
        raise ValueError(
            "Sum of input lengths does not equal the length of the input dataset!"
        )

    indices = paddle.randperm(sum(lengths))
    return [
        Subset(dataset, indices[offset - length:offset])
        for offset, length in zip(_accumulate(lengths), lengths)
    ]


setattr(paddle.io, "random_split", random_split)


class DataLoader(paddle.io.DataLoader):
    def __init__(self,
                 dataset,
                 batch_size=1,
                 shuffle=False,
                 sampler=None,
                 batch_sampler=None,
                 num_workers=0,
                 collate_fn=None,
                 pin_memory=False,
                 drop_last=False,
                 timeout=0,
                 worker_init_fn=None,
                 multiprocessing_context=None,
                 generator=None):
        if isinstance(dataset[0], (tuple, list)):
            return_list = True
        else:
            return_list = False
        return_list = True
        super().__init__(
            dataset,
            feed_list=None,
            places=None,
            return_list=return_list,
            batch_sampler=batch_sampler,
            batch_size=batch_size,
            shuffle=shuffle,
            drop_last=drop_last,
            collate_fn=collate_fn,
            num_workers=num_workers,
            use_buffer_reader=True,
            use_shared_memory=False,
            timeout=timeout,
            worker_init_fn=worker_init_fn)
        if sampler is not None:
            seld.batch_sampler.sampler = sampler


class DistributedSampler(paddle.io.DistributedBatchSampler):
    def __init__(self,
                 dataset,
                 num_replicas=None,
                 rank=None,
                 shuffle=True,
                 seed=0,
                 drop_last=False):
        super().__init__(
            dataset=dataset,
            batch_size=1,
            num_replicas=num_replicas,
            rank=rank,
            shuffle=shuffle,
            drop_last=drop_last)



# Copyright (c) MLTALKER. All rights reserved.
import sys
sys.path.append('/root/paddlejob/workspace/env_run/portrait/antgo')
sys.path.append('/root/paddlejob/workspace/env_run/portrait/InterHand26M/common')
sys.path.append('/root/paddlejob/workspace/env_run/portrait/InterHand26M/main')
import argparse
import copy
import os
import os.path as osp
import time
import warnings
import numpy as np

import torch
import torch.distributed as dist
from antgo.framework.helper.utils import Config, DictAction
from antgo.framework.helper.runner import get_dist_info, init_dist
from antgo.framework.helper.utils import get_git_hash
from antgo.framework.helper.utils import (collect_env,update_data_root)
from antgo.framework.helper.utils import (collect_env, get_logger, update_data_root)
from antgo.framework.helper.utils.util_distribution import get_device
from antgo.framework.helper.utils.setup_env import *
from antgo.framework.helper.runner.utils import *
from antgo.framework.helper.models.builder import *
from antgo.framework.helper.apis.train import *
from antgo.framework.helper.dataset import *
from antgo.framework.helper.tools.util import *
from antgo.framework.helper.runner.checkpoint import load_checkpoint
from antgo.framework.helper.cnn import fuse_conv_bn
from antgo.framework.helper.runner.test import multi_gpu_test, single_gpu_test
from antgo.framework.helper.runner.builder import *
from antgo.framework.helper.utils.util_distribution import get_device
import json


def main():
    args = parse_args()
    cfg = Config.fromfile(args.config)
    print(f'checkpoint {args.checkpoint}')
    # set multi-process settings
    # 如 opencv_num_threads, OMP_NUM_THREADS, MKL_NUM_THREADS
    setup_multi_processes(cfg)

    # set cudnn_benchmark
    if cfg.get('cudnn_benchmark', False):
        torch.backends.cudnn.benchmark = True

    # work_dir is determined in this priority: CLI > segment in file > filename
    if args.work_dir is not None and args.work_dir != '':
        # update configs according to CLI args if args.work_dir is not None
        cfg.work_dir = args.work_dir
    elif cfg.get('work_dir', None) is None:
        # use config filename as default work_dir if cfg.work_dir is None
        cfg.work_dir = osp.join('./work_dirs',
                                osp.splitext(osp.basename(args.config))[0])

    # 非分布式下，仅支持单卡
    cfg.gpu_ids = [args.gpu_id]

    # init distributed env first, since logger depends on the dist info.
    if args.launcher == 'none':
        distributed = False
    else:
        distributed = True
        init_dist(**cfg.get('dist_params', {}))

    test_dataloader_default_args = dict(
        samples_per_gpu=1, workers_per_gpu=2, dist=distributed, shuffle=False)

    test_loader_cfg = {
        **test_dataloader_default_args,
        **cfg.data.get('test_dataloader', {})
    }

    rank, _ = get_dist_info()
    # allows not to create
    if cfg.work_dir is not None and rank == 0:
        if not os.path.exists(osp.abspath(cfg.work_dir)):
            os.makedirs(osp.abspath(cfg.work_dir))
        timestamp = time.strftime('%Y-%m-%dx%H-%M-%S', time.localtime())
        json_file = osp.join(cfg.work_dir, f'eval_{timestamp}.json')

    # build the dataloader
    dataset = build_dataset(cfg.data.test)
    data_loader = build_dataloader(dataset, **test_loader_cfg)

    # build the model and load checkpoint
    cfg.model.train_cfg = None
    if args.ext_module != '':
        # 导入ext module
        print(f'load ext model from {args.ext_module}')
        load_extmodule(args.ext_module)
    
    model = \
        build_model(cfg.model, test_cfg=cfg.get('test_cfg', None))
    checkpoint = load_checkpoint(model, args.checkpoint, map_location='cpu')
    if args.fuse_conv_bn:
        model = fuse_conv_bn(model)

    cfg.device = get_device()
    if not distributed:
        model = build_dp(model, cfg.get('device', 'cuda'), device_ids=cfg.gpu_ids)
        outputs = single_gpu_test(model, data_loader)
    else:
        model = build_ddp(
            model,
            cfg.get('device', 'cuda'),
            device_ids=[int(os.environ['LOCAL_RANK'])],
            broadcast_buffers=False)
        outputs = multi_gpu_test(model, data_loader)

    rank, _ = get_dist_info()
    if rank == 0:
        metric_func = None
        if 'metric' in cfg:
            metric_func = build_measure(cfg['metric'])         

        eval_kwargs = cfg.get('evaluation', {}).copy()  
        if metric_func is None:
            metric = dataset.evaluate(
                outputs, **eval_kwargs)
        else:
            gts = dataset.get_ann_info(None)
            if gts is None:
                gts = []
                for gt_i in len(dataset):
                    gts.append(dataset.sample(gt_i))
            metric = metric_func(outputs, gts)

        print(metric)
        metric_dict = dict(config=args.config, metric=metric)
        if cfg.work_dir is not None and rank == 0:
            json.dump(metric_dict, json_file)

if __name__ == '__main__':
    main()



import torch
import torch.nn as nn
from antgo.framework.helper.cnn.bricks import *
from antgo.framework.helper.models.builder import HEADS, MODELS, build_loss
from antgo.framework.helper.runner import BaseModule
import torch.nn.functional as F


class ASPPModule(nn.ModuleList):
    """Atrous Spatial Pyramid Pooling (ASPP) Module.
    Args:
        dilations (tuple[int]): Dilation rate of each layer.
        in_channels (int): Input channels.
        channels (int): Channels after modules, before conv_seg.
        conv_cfg (dict|None): Config of conv layers.
        norm_cfg (dict|None): Config of norm layers.
        act_cfg (dict): Config of activation layers.
    """

    def __init__(self, dilations, in_channels, channels, conv_cfg, norm_cfg,
                 act_cfg):
        super().__init__()
        self.dilations = dilations
        self.in_channels = in_channels
        self.channels = channels
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg
        for dilation in dilations:
            self.append(
                ConvModule(
                    self.in_channels,
                    self.channels,
                    1 if dilation == 1 else 3,
                    dilation=dilation,
                    padding=0 if dilation == 1 else dilation,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg,
                    act_cfg=self.act_cfg))

    def forward(self, x):
        """Forward function."""
        aspp_outs = []
        for aspp_module in self:
            aspp_outs.append(aspp_module(x))

        return aspp_outs


@HEADS.register_module()
class ASPPHead(BaseModule):
    """Rethinking Atrous Convolution for Semantic Image Segmentation.
    This head is the implementation of `DeepLabV3
    <https://arxiv.org/abs/1706.05587>`_.
    Args:
        dilations (tuple[int]): Dilation rates for ASPP module.
            Default: (1, 6, 12, 18).
    """

    def __init__(self, 
                 dilations=(1, 6, 12, 18), 
                 in_channels=160, 
                 channels=32, 
                 out_channels=2, 
                 dropout_ratio=0.1,
                 loss_decode=dict(
                     type='CrossEntropyLoss',
                     use_sigmoid=False,
                     loss_weight=1.0),
                 align_corners=False,
                 ignore_index=None,
                 **kwargs):
        super().__init__(init_cfg=kwargs.get('init_cfg', None))
        assert isinstance(dilations, (list, tuple))
        self.dilations = dilations
        self.in_channels = in_channels
        self.channels = channels
        self.out_channels = out_channels
        self.align_corners = align_corners
        self.loss_decode = build_loss(loss_decode)
        self.ignore_index = ignore_index
        
        self.image_pool = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            ConvModule(
                self.in_channels,
                self.channels,
                1,
                conv_cfg=None,
                norm_cfg=dict(type='BN'),
                act_cfg=dict(type='ReLU')))
        self.aspp_modules = ASPPModule(
            dilations,
            self.in_channels,
            self.channels,
            conv_cfg=None,
            norm_cfg=dict(type='BN'),
            act_cfg=dict(type='ReLU'))
        self.bottleneck = ConvModule(
            (len(dilations) + 1) * self.channels,
            self.channels,
            3,
            padding=1,
            conv_cfg=None,
            norm_cfg=dict(type='BN'),
            act_cfg=dict(type='ReLU'))
        
        self.conv_seg = nn.Conv2d(channels, self.out_channels, kernel_size=1)
        if dropout_ratio > 0:
            self.dropout = nn.Dropout2d(dropout_ratio)
        else:
            self.dropout = None

    def _forward_feature(self, inputs):
        """Forward function for feature maps before classifying each pixel with
        ``self.cls_seg`` fc.
        Args:
            inputs (list[Tensor]): List of multi-level img features.
        Returns:
            feats (Tensor): A tensor of shape (batch_size, self.channels,
                H, W) which is feature map for last layer of decoder head.
        """
        x = inputs[-1]
        aspp_outs = [
            F.interpolate(self.image_pool(x), x.size()[2:], None, mode='bilinear', align_corners=self.align_corners)
        ]
        aspp_outs.extend(self.aspp_modules(x))
        aspp_outs = torch.cat(aspp_outs, dim=1)
        feats = self.bottleneck(aspp_outs)
        return feats

    def loss_by_feat(self, seg_logits, seg_labels):
        """Compute segmentation loss.
        Args:
            seg_logits (Tensor): The output from decode head forward function.
            batch_data_samples (List[:obj:`SegDataSample`]): The seg
                data samples. It usually includes information such
                as `metainfo` and `gt_sem_seg`.
        Returns:
            dict[str, Tensor]: a dictionary of loss components
        """

        loss = dict()
        seg_logits = F.interpolate(seg_logits, seg_labels.shape[2:], mode='bilinear', align_corners=self.align_corners)
        seg_labels = seg_labels.squeeze(1)

        if not isinstance(self.loss_decode, nn.ModuleList):
            losses_decode = [self.loss_decode]
        else:
            losses_decode = self.loss_decode

        for loss_decode in losses_decode:
            loss['loss_seg'] = loss_decode(
                seg_logits,
                seg_labels,
                ignore_index=self.ignore_index)

        return loss
    
    def cls_seg(self, feat):
        """Classify each pixel."""
        if self.dropout is not None:
            feat = self.dropout(feat)
        output = self.conv_seg(feat)
        return output

    def forward(self, inputs):
        """Forward function."""
        output = self._forward_feature(inputs)
        output = self.cls_seg(output)
        return output

    def loss(self, seg_logits, seg_labels):
        loss_dict = dict()
        seg_logits = F.interpolate(seg_logits, seg_labels.shape[2:], mode='bilinear', align_corners=self.align_corners)
        seg_labels = seg_labels.squeeze(1)

        if not isinstance(self.loss_decode, nn.ModuleList):
            losses_decode = [self.loss_decode]
        else:
            losses_decode = self.loss_decode

        for loss_decode in losses_decode:
            loss_dict['loss_seg'] = loss_decode(
                seg_logits,
                seg_labels,
                ignore_index=self.ignore_index)

        return loss_dict
    
    def simple_test(self, x):
        x = self.forward(x)
        x = torch.softmax(x, 1)
        return x


from ..config import config
from ..utils import makedirs
from .build_hash import path_hash
import os


def get_virtual_dirname(path):
    if config.BUILD_IN_LOCAL_PATH:
        return path
    assert os.path.isdir(path), path
    dirname, basename = os.path.split(path)
    # hash dirname
    hash_dirname = path_hash(dirname)
    new_path = os.path.normpath(os.path.join(
        config.BUILD_PATH, 'build', '{}_{}'.format(basename, hash_dirname)))
    makedirs(new_path, exist_ok=True)
    tag_fname = os.path.join(new_path, 'ORIGINAL_PATH')
    if not os.path.exists(tag_fname):
        with open(tag_fname, 'w') as fout:
            fout.write(os.path.abspath(path))
    return new_path


def change_exts(lst, rules):
    res = []
    mappings = dict(rules)
    for name in lst:
        sp = os.path.splitext(name)
        if sp[1] and sp[1][0] == '.':
            ext = sp[1][1:]
            if ext in mappings:
                new_ext = mappings[ext]
                name = sp[0] + '.' + new_ext
        res.append(name)
    return res


def add_path(path, files):
    return list(map(lambda x: os.path.join(path, x), files))



# Copyright (c) OpenMMLab. All rights reserved.
from .hook import HOOKS, Hook
from .lr_updater import annealing_cos, annealing_linear, format_param


class MomentumUpdaterHook(Hook):

    def __init__(self,
                 by_epoch=True,
                 warmup=None,
                 warmup_iters=0,
                 warmup_ratio=0.9):
        # validate the "warmup" argument
        if warmup is not None:
            if warmup not in ['constant', 'linear', 'exp']:
                raise ValueError(
                    f'"{warmup}" is not a supported type for warming up, valid'
                    ' types are "constant" and "linear"')
        if warmup is not None:
            assert warmup_iters > 0, \
                '"warmup_iters" must be a positive integer'
            assert 0 < warmup_ratio <= 1.0, \
                '"warmup_momentum" must be in range (0,1]'

        self.by_epoch = by_epoch
        self.warmup = warmup
        self.warmup_iters = warmup_iters
        self.warmup_ratio = warmup_ratio

        self.base_momentum = []  # initial momentum for all param groups
        self.regular_momentum = [
        ]  # expected momentum if no warming up is performed

    def _set_momentum(self, runner, momentum_groups):
        if isinstance(runner.optimizer, dict):
            for k, optim in runner.optimizer.items():
                for param_group, mom in zip(optim.param_groups,
                                            momentum_groups[k]):
                    if 'momentum' in param_group.keys():
                        param_group['momentum'] = mom
                    elif 'betas' in param_group.keys():
                        param_group['betas'] = (mom, param_group['betas'][1])
        else:
            for param_group, mom in zip(runner.optimizer.param_groups,
                                        momentum_groups):
                if 'momentum' in param_group.keys():
                    param_group['momentum'] = mom
                elif 'betas' in param_group.keys():
                    param_group['betas'] = (mom, param_group['betas'][1])

    def get_momentum(self, runner, base_momentum):
        raise NotImplementedError

    def get_regular_momentum(self, runner):
        if isinstance(runner.optimizer, dict):
            momentum_groups = {}
            for k in runner.optimizer.keys():
                _momentum_group = [
                    self.get_momentum(runner, _base_momentum)
                    for _base_momentum in self.base_momentum[k]
                ]
                momentum_groups.update({k: _momentum_group})
            return momentum_groups
        else:
            return [
                self.get_momentum(runner, _base_momentum)
                for _base_momentum in self.base_momentum
            ]

    def get_warmup_momentum(self, cur_iters):

        def _get_warmup_momentum(cur_iters, regular_momentum):
            if self.warmup == 'constant':
                warmup_momentum = [
                    _momentum / self.warmup_ratio
                    for _momentum in regular_momentum
                ]
            elif self.warmup == 'linear':
                k = (1 - cur_iters / self.warmup_iters) * (1 -
                                                           self.warmup_ratio)
                warmup_momentum = [
                    _momentum / (1 - k) for _momentum in regular_momentum
                ]
            elif self.warmup == 'exp':
                k = self.warmup_ratio**(1 - cur_iters / self.warmup_iters)
                warmup_momentum = [
                    _momentum / k for _momentum in regular_momentum
                ]
            return warmup_momentum

        if isinstance(self.regular_momentum, dict):
            momentum_groups = {}
            for key, regular_momentum in self.regular_momentum.items():
                momentum_groups[key] = _get_warmup_momentum(
                    cur_iters, regular_momentum)
            return momentum_groups
        else:
            return _get_warmup_momentum(cur_iters, self.regular_momentum)

    def before_run(self, runner):
        # NOTE: when resuming from a checkpoint,
        # if 'initial_momentum' is not saved,
        # it will be set according to the optimizer params
        if isinstance(runner.optimizer, dict):
            self.base_momentum = {}
            for k, optim in runner.optimizer.items():
                for group in optim.param_groups:
                    if 'momentum' in group.keys():
                        group.setdefault('initial_momentum', group['momentum'])
                    else:
                        group.setdefault('initial_momentum', group['betas'][0])
                _base_momentum = [
                    group['initial_momentum'] for group in optim.param_groups
                ]
                self.base_momentum.update({k: _base_momentum})
        else:
            for group in runner.optimizer.param_groups:
                if 'momentum' in group.keys():
                    group.setdefault('initial_momentum', group['momentum'])
                else:
                    group.setdefault('initial_momentum', group['betas'][0])
            self.base_momentum = [
                group['initial_momentum']
                for group in runner.optimizer.param_groups
            ]

    def before_train_epoch(self, runner):
        if not self.by_epoch:
            return
        self.regular_momentum = self.get_regular_momentum(runner)
        self._set_momentum(runner, self.regular_momentum)

    def before_train_iter(self, runner):
        cur_iter = runner.iter
        if not self.by_epoch:
            self.regular_momentum = self.get_regular_momentum(runner)
            if self.warmup is None or cur_iter >= self.warmup_iters:
                self._set_momentum(runner, self.regular_momentum)
            else:
                warmup_momentum = self.get_warmup_momentum(cur_iter)
                self._set_momentum(runner, warmup_momentum)
        elif self.by_epoch:
            if self.warmup is None or cur_iter > self.warmup_iters:
                return
            elif cur_iter == self.warmup_iters:
                self._set_momentum(runner, self.regular_momentum)
            else:
                warmup_momentum = self.get_warmup_momentum(cur_iter)
                self._set_momentum(runner, warmup_momentum)


@HOOKS.register_module()
class StepMomentumUpdaterHook(MomentumUpdaterHook):
    """Step momentum scheduler with min value clipping.

    Args:
        step (int | list[int]): Step to decay the momentum. If an int value is
            given, regard it as the decay interval. If a list is given, decay
            momentum at these steps.
        gamma (float, optional): Decay momentum ratio. Default: 0.5.
        min_momentum (float, optional): Minimum momentum value to keep. If
            momentum after decay is lower than this value, it will be clipped
            accordingly. If None is given, we don't perform lr clipping.
            Default: None.
    """

    def __init__(self, step, gamma=0.5, min_momentum=None, **kwargs):
        if isinstance(step, list):
            assert all([s > 0 for s in step])
        elif isinstance(step, int):
            assert step > 0
        else:
            raise TypeError('"step" must be a list or integer')
        self.step = step
        self.gamma = gamma
        self.min_momentum = min_momentum
        super(StepMomentumUpdaterHook, self).__init__(**kwargs)

    def get_momentum(self, runner, base_momentum):
        progress = runner.epoch if self.by_epoch else runner.iter

        # calculate exponential term
        if isinstance(self.step, int):
            exp = progress // self.step
        else:
            exp = len(self.step)
            for i, s in enumerate(self.step):
                if progress < s:
                    exp = i
                    break

        momentum = base_momentum * (self.gamma**exp)
        if self.min_momentum is not None:
            # clip to a minimum value
            momentum = max(momentum, self.min_momentum)
        return momentum


@HOOKS.register_module()
class CosineAnnealingMomentumUpdaterHook(MomentumUpdaterHook):
    """Cosine annealing LR Momentum decays the Momentum of each parameter group
    linearly.

    Args:
        min_momentum (float, optional): The minimum momentum. Default: None.
        min_momentum_ratio (float, optional): The ratio of minimum momentum to
            the base momentum. Either `min_momentum` or `min_momentum_ratio`
            should be specified. Default: None.
    """

    def __init__(self, min_momentum=None, min_momentum_ratio=None, **kwargs):
        assert (min_momentum is None) ^ (min_momentum_ratio is None)
        self.min_momentum = min_momentum
        self.min_momentum_ratio = min_momentum_ratio
        super(CosineAnnealingMomentumUpdaterHook, self).__init__(**kwargs)

    def get_momentum(self, runner, base_momentum):
        if self.by_epoch:
            progress = runner.epoch
            max_progress = runner.max_epochs
        else:
            progress = runner.iter
            max_progress = runner.max_iters
        if self.min_momentum_ratio is not None:
            target_momentum = base_momentum * self.min_momentum_ratio
        else:
            target_momentum = self.min_momentum
        return annealing_cos(base_momentum, target_momentum,
                             progress / max_progress)


@HOOKS.register_module()
class LinearAnnealingMomentumUpdaterHook(MomentumUpdaterHook):
    """Linear annealing LR Momentum decays the Momentum of each parameter group
    linearly.

    Args:
        min_momentum (float, optional): The minimum momentum. Default: None.
        min_momentum_ratio (float, optional): The ratio of minimum momentum to
            the base momentum. Either `min_momentum` or `min_momentum_ratio`
            should be specified. Default: None.
    """

    def __init__(self, min_momentum=None, min_momentum_ratio=None, **kwargs):
        assert (min_momentum is None) ^ (min_momentum_ratio is None)
        self.min_momentum = min_momentum
        self.min_momentum_ratio = min_momentum_ratio
        super(LinearAnnealingMomentumUpdaterHook, self).__init__(**kwargs)

    def get_momentum(self, runner, base_momentum):
        if self.by_epoch:
            progress = runner.epoch
            max_progress = runner.max_epochs
        else:
            progress = runner.iter
            max_progress = runner.max_iters
        if self.min_momentum_ratio is not None:
            target_momentum = base_momentum * self.min_momentum_ratio
        else:
            target_momentum = self.min_momentum
        return annealing_linear(base_momentum, target_momentum,
                                progress / max_progress)


@HOOKS.register_module()
class CyclicMomentumUpdaterHook(MomentumUpdaterHook):
    """Cyclic momentum Scheduler.

    Implement the cyclical momentum scheduler policy described in
    https://arxiv.org/pdf/1708.07120.pdf

    This momentum scheduler usually used together with the CyclicLRUpdater
    to improve the performance in the 3D detection area.

    Args:
        target_ratio (tuple[float]): Relative ratio of the lowest momentum and
            the highest momentum to the initial momentum.
        cyclic_times (int): Number of cycles during training
        step_ratio_up (float): The ratio of the increasing process of momentum
            in  the total cycle.
        by_epoch (bool): Whether to update momentum by epoch.
        anneal_strategy (str, optional): {'cos', 'linear'}
            Specifies the annealing strategy: 'cos' for cosine annealing,
            'linear' for linear annealing. Default: 'cos'.
        gamma (float, optional): Cycle decay ratio. Default: 1.
            It takes values in the range (0, 1]. The difference between the
            maximum learning rate and the minimum learning rate decreases
            periodically when it is less than 1. `New in version 1.4.4.`
    """

    def __init__(self,
                 by_epoch=False,
                 target_ratio=(0.85 / 0.95, 1),
                 cyclic_times=1,
                 step_ratio_up=0.4,
                 anneal_strategy='cos',
                 gamma=1,
                 **kwargs):
        if isinstance(target_ratio, float):
            target_ratio = (target_ratio, target_ratio / 1e5)
        elif isinstance(target_ratio, tuple):
            target_ratio = (target_ratio[0], target_ratio[0] / 1e5) \
                if len(target_ratio) == 1 else target_ratio
        else:
            raise ValueError('target_ratio should be either float '
                             f'or tuple, got {type(target_ratio)}')

        assert len(target_ratio) == 2, \
            '"target_ratio" must be list or tuple of two floats'
        assert 0 <= step_ratio_up < 1.0, \
            '"step_ratio_up" must be in range [0,1)'

        self.target_ratio = target_ratio
        self.cyclic_times = cyclic_times
        self.step_ratio_up = step_ratio_up
        self.gamma = gamma
        self.momentum_phases = []  # init momentum_phases

        if anneal_strategy not in ['cos', 'linear']:
            raise ValueError('anneal_strategy must be one of "cos" or '
                             f'"linear", instead got {anneal_strategy}')
        elif anneal_strategy == 'cos':
            self.anneal_func = annealing_cos
        elif anneal_strategy == 'linear':
            self.anneal_func = annealing_linear
        # currently only support by_epoch=False
        assert not by_epoch, \
            'currently only support "by_epoch" = False'
        super(CyclicMomentumUpdaterHook, self).__init__(by_epoch, **kwargs)

    def before_run(self, runner):
        super(CyclicMomentumUpdaterHook, self).before_run(runner)
        # initiate momentum_phases
        # total momentum_phases are separated as up and down
        max_iter_per_phase = runner.max_iters // self.cyclic_times
        iter_up_phase = int(self.step_ratio_up * max_iter_per_phase)
        self.max_iter_per_phase = max_iter_per_phase
        self.momentum_phases.append(
            [0, iter_up_phase, 1, self.target_ratio[0]])
        self.momentum_phases.append([
            iter_up_phase, max_iter_per_phase, self.target_ratio[0],
            self.target_ratio[1]
        ])

    def get_momentum(self, runner, base_momentum):
        curr_iter = runner.iter % self.max_iter_per_phase
        curr_cycle = runner.iter // self.max_iter_per_phase
        scale = self.gamma**curr_cycle
        for (start_iter, end_iter, start_ratio, end_ratio) \
                in self.momentum_phases:
            if start_iter <= curr_iter < end_iter:
                # Apply cycle scaling to gradually reduce the difference
                # between max_momentum and base momentum. The target end_ratio
                # can be expressed as:
                # end_ratio = (base_momentum + scale * \
                # (max_momentum - base_momentum)) / base_momentum
                # iteration: 0-iter_up_phase:
                if start_iter == 0:
                    end_ratio = 1 - scale + end_ratio * scale
                # iteration: iter_up_phase-self.max_iter_per_phase
                else:
                    start_ratio = 1 - scale + start_ratio * scale
                progress = curr_iter - start_iter
                return self.anneal_func(base_momentum * start_ratio,
                                        base_momentum * end_ratio,
                                        progress / (end_iter - start_iter))


@HOOKS.register_module()
class OneCycleMomentumUpdaterHook(MomentumUpdaterHook):
    """OneCycle momentum Scheduler.

    This momentum scheduler usually used together with the OneCycleLrUpdater
    to improve the performance.

    Args:
        base_momentum (float or list): Lower momentum boundaries in the cycle
            for each parameter group. Note that momentum is cycled inversely
            to learning rate; at the peak of a cycle, momentum is
            'base_momentum' and learning rate is 'max_lr'.
            Default: 0.85
        max_momentum (float or list): Upper momentum boundaries in the cycle
            for each parameter group. Functionally,
            it defines the cycle amplitude (max_momentum - base_momentum).
            Note that momentum is cycled inversely
            to learning rate; at the start of a cycle, momentum is
            'max_momentum' and learning rate is 'base_lr'
            Default: 0.95
        pct_start (float): The percentage of the cycle (in number of steps)
            spent increasing the learning rate.
            Default: 0.3
        anneal_strategy (str): {'cos', 'linear'}
            Specifies the annealing strategy: 'cos' for cosine annealing,
            'linear' for linear annealing.
            Default: 'cos'
        three_phase (bool): If three_phase is True, use a third phase of the
            schedule to annihilate the learning rate according to
            final_div_factor instead of modifying the second phase (the first
            two phases will be symmetrical about the step indicated by
            pct_start).
            Default: False
    """

    def __init__(self,
                 base_momentum=0.85,
                 max_momentum=0.95,
                 pct_start=0.3,
                 anneal_strategy='cos',
                 three_phase=False,
                 **kwargs):
        # validate by_epoch, currently only support by_epoch=False
        if 'by_epoch' not in kwargs:
            kwargs['by_epoch'] = False
        else:
            assert not kwargs['by_epoch'], \
                'currently only support "by_epoch" = False'
        if not isinstance(base_momentum, (float, list, dict)):
            raise ValueError('base_momentum must be the type among of float,'
                             'list or dict.')
        self._base_momentum = base_momentum
        if not isinstance(max_momentum, (float, list, dict)):
            raise ValueError('max_momentum must be the type among of float,'
                             'list or dict.')
        self._max_momentum = max_momentum
        # validate pct_start
        if pct_start < 0 or pct_start > 1 or not isinstance(pct_start, float):
            raise ValueError('Expected float between 0 and 1 pct_start, but '
                             f'got {pct_start}')
        self.pct_start = pct_start
        # validate anneal_strategy
        if anneal_strategy not in ['cos', 'linear']:
            raise ValueError('anneal_strategy must by one of "cos" or '
                             f'"linear", instead got {anneal_strategy}')
        elif anneal_strategy == 'cos':
            self.anneal_func = annealing_cos
        elif anneal_strategy == 'linear':
            self.anneal_func = annealing_linear
        self.three_phase = three_phase
        self.momentum_phases = []  # init momentum_phases
        super(OneCycleMomentumUpdaterHook, self).__init__(**kwargs)

    def before_run(self, runner):
        if isinstance(runner.optimizer, dict):
            for k, optim in runner.optimizer.items():
                if ('momentum' not in optim.defaults
                        and 'betas' not in optim.defaults):
                    raise ValueError('optimizer must support momentum with'
                                     'option enabled')
                self.use_beta1 = 'betas' in optim.defaults
                _base_momentum = format_param(k, optim, self._base_momentum)
                _max_momentum = format_param(k, optim, self._max_momentum)
                for group, b_momentum, m_momentum in zip(
                        optim.param_groups, _base_momentum, _max_momentum):
                    if self.use_beta1:
                        _, beta2 = group['betas']
                        group['betas'] = (m_momentum, beta2)
                    else:
                        group['momentum'] = m_momentum
                    group['base_momentum'] = b_momentum
                    group['max_momentum'] = m_momentum
        else:
            optim = runner.optimizer
            if ('momentum' not in optim.defaults
                    and 'betas' not in optim.defaults):
                raise ValueError('optimizer must support momentum with'
                                 'option enabled')
            self.use_beta1 = 'betas' in optim.defaults
            k = type(optim).__name__
            _base_momentum = format_param(k, optim, self._base_momentum)
            _max_momentum = format_param(k, optim, self._max_momentum)
            for group, b_momentum, m_momentum in zip(optim.param_groups,
                                                     _base_momentum,
                                                     _max_momentum):
                if self.use_beta1:
                    _, beta2 = group['betas']
                    group['betas'] = (m_momentum, beta2)
                else:
                    group['momentum'] = m_momentum
                group['base_momentum'] = b_momentum
                group['max_momentum'] = m_momentum

        if self.three_phase:
            self.momentum_phases.append({
                'end_iter':
                float(self.pct_start * runner.max_iters) - 1,
                'start_momentum':
                'max_momentum',
                'end_momentum':
                'base_momentum'
            })
            self.momentum_phases.append({
                'end_iter':
                float(2 * self.pct_start * runner.max_iters) - 2,
                'start_momentum':
                'base_momentum',
                'end_momentum':
                'max_momentum'
            })
            self.momentum_phases.append({
                'end_iter': runner.max_iters - 1,
                'start_momentum': 'max_momentum',
                'end_momentum': 'max_momentum'
            })
        else:
            self.momentum_phases.append({
                'end_iter':
                float(self.pct_start * runner.max_iters) - 1,
                'start_momentum':
                'max_momentum',
                'end_momentum':
                'base_momentum'
            })
            self.momentum_phases.append({
                'end_iter': runner.max_iters - 1,
                'start_momentum': 'base_momentum',
                'end_momentum': 'max_momentum'
            })

    def _set_momentum(self, runner, momentum_groups):
        if isinstance(runner.optimizer, dict):
            for k, optim in runner.optimizer.items():
                for param_group, mom in zip(optim.param_groups,
                                            momentum_groups[k]):
                    if 'momentum' in param_group.keys():
                        param_group['momentum'] = mom
                    elif 'betas' in param_group.keys():
                        param_group['betas'] = (mom, param_group['betas'][1])
        else:
            for param_group, mom in zip(runner.optimizer.param_groups,
                                        momentum_groups):
                if 'momentum' in param_group.keys():
                    param_group['momentum'] = mom
                elif 'betas' in param_group.keys():
                    param_group['betas'] = (mom, param_group['betas'][1])

    def get_momentum(self, runner, param_group):
        curr_iter = runner.iter
        start_iter = 0
        for i, phase in enumerate(self.momentum_phases):
            end_iter = phase['end_iter']
            if curr_iter <= end_iter or i == len(self.momentum_phases) - 1:
                pct = (curr_iter - start_iter) / (end_iter - start_iter)
                momentum = self.anneal_func(
                    param_group[phase['start_momentum']],
                    param_group[phase['end_momentum']], pct)
                break
            start_iter = end_iter
        return momentum

    def get_regular_momentum(self, runner):
        if isinstance(runner.optimizer, dict):
            momentum_groups = {}
            for k, optim in runner.optimizer.items():
                _momentum_group = [
                    self.get_momentum(runner, param_group)
                    for param_group in optim.param_groups
                ]
                momentum_groups.update({k: _momentum_group})
            return momentum_groups
        else:
            momentum_groups = []
            for param_group in runner.optimizer.param_groups:
                momentum_groups.append(self.get_momentum(runner, param_group))
            return momentum_groups



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 17:54
# @File    : base.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from abc import abstractmethod
from abc import ABC
from enum import Enum
from enum import Flag, auto

# NotShareable:
#    Stateful & reusable operators.

# NotRusable:
#    Stateful & not reusable operators.

# Shareable:
#    Stateless operators

SharedType = Enum('SharedType', ('NotShareable', 'NotReusable', 'Shareable'))

class OperatorFlag(Flag):
    EMPTYFLAG = auto()
    STATELESS = auto()
    REUSEABLE = auto()

class Operator(ABC):
    """
    Operator base class, implements __init__ and __call__,

    Examples:
        class AddOperator(Operator):
            def __init__(self, factor: int):
                self._factor = factor

            def __call__(self, num) -> NamedTuple("Outputs", [("sum", int)]):
                Outputs = NamedTuple("Outputs", [("sum", int)])
                return Outputs(self._factor + num)
    """

    @abstractmethod
    def __init__(self):
        """
        Init operators, before a graph starts, the framework will call Operator __init__ function.

        Args:

        Raises:
            An exception during __init__ can terminate the graph run.
        """
        self._key = ''

    @abstractmethod
    def __call__(self):
        """
        The framework calls __call__ function repeatedly for every input data.

        Args:

        Returns:

        Raises:
            An exception during __init__ can terminate the graph run.
        """
        raise NotImplementedError

    @property
    def key(self):
        return self._key

    @property
    def shared_type(self):
        return SharedType.NotShareable

    @key.setter
    def key(self, value):
        self._key = value

    @property
    def flag(self):
        return OperatorFlag.STATELESS|OperatorFlag.REUSEABLE


class NNOperator(Operator):
    """
    Neural Network related operators that involve machine learning frameworks.

    Args:
        framework (`str`):
            The framework to apply.
    """

    def __init__(self, framework: str = 'pytorch'):
        super().__init__()
        self._framework = framework
        self.operator_name = type(self).__name__
        self.model = None
        self.model_card = None
        self._trainer = None

    @property
    def flag(self):
        return OperatorFlag.STATELESS|OperatorFlag.REUSEABLE

    @property
    def framework(self):
        return self._framework

    @framework.setter
    def framework(self, framework: str):
        self._framework = framework

    def save_model(self):
        """
        Save model to local.
        """
        raise NotImplementedError

    def supported_model_names(self):
        """
        Return a list of supported model names.
        """
        raise NotImplementedError

    def train(self,
              training_config=None,
              train_dataset=None,
              eval_dataset=None,
              resume_checkpoint_path=None,
              **kwargs):
        """
        Start to train an operators.

        Args:
            training_config (`TrainingConfig`):
                The config of this trainer.
            train_dataset (`Union[Dataset, TowheeDataSet]`):
                Training dataset.
            eval_dataset (`Union[Dataset, TowheeDataSet]`):
                Evaluate dataset.
            resume_checkpoint_path (`str`):
                If resuming training, pass into the path.
            **kwargs (`Any`):
                Keyword Args.
        """
        self.setup_trainer(training_config, train_dataset, eval_dataset, **kwargs)
        self.trainer.train(resume_checkpoint_path)

    @property
    def trainer(self) -> 'Trainer':
        if self._trainer is None:
            self.setup_trainer()
        return self._trainer

    @trainer.setter
    def trainer(self, trainer):
        self._trainer = trainer

    def setup_trainer(self, training_config=None,
                      train_dataset=None,
                      eval_dataset=None,
                      train_dataloader=None,
                      eval_dataloader=None,
                      model_card=None
                      ):
        """
        Set up the trainer instance in operators before training and set trainer parameters.
        Args:
            training_config (`TrainingConfig`):
                The config of this trainer.
            train_dataset (`Union[Dataset, TowheeDataSet]`):
                Training dataset.
            eval_dataset (`Union[Dataset, TowheeDataSet]`):
                Evaluate dataset.
            train_dataloader (`Union[DataLoader, Iterable]`):
                If specified, `Trainer` will use it to load training data.
                Otherwise, `Trainer` will build dataloader from train_dataset.
            eval_dataloader (`Union[DataLoader, Iterable]`):
                If specified, `Trainer` will use it to load evaluate data.
                Otherwise, `Trainer` will build dataloader from train_dataset.
            model_card (`ModelCard`):
                Model card contains the training informations.
        Returns:

        """
        from towhee.trainer.trainer import Trainer  # pylint: disable=import-outside-toplevel
        if self._trainer is None:
            self._trainer = Trainer(self.model,
                                    training_config,
                                    train_dataset,
                                    eval_dataset,
                                    train_dataloader,
                                    eval_dataloader,
                                    model_card)
        else:
            if training_config is not None:
                self._trainer.configs = training_config
            if train_dataset is not None:
                self._trainer.train_dataset = train_dataset
            if eval_dataset is not None:
                self._trainer.eval_dataset = eval_dataset
            if train_dataloader is not None:
                self._trainer.train_dataloader = train_dataloader
            if eval_dataloader is not None:
                self._trainer.eval_dataloader = eval_dataloader
            if model_card is not None:
                self._trainer.model_card = model_card

    def load(self, path: str = None):
        """
        Load the model checkpoint into an operators.

        Args:
            path (`str`):
                The folder path containing the model's checkpoints.
        """
        self.trainer.load(path)

    def save(self, path: str, overwrite: bool = True):
        """
        Save the model checkpoint into the path.

        Args:
            path (`str`):
                The folder path containing the model's checkpoints.
            overwrite (`bool`):
                If True, it will overwrite the same name path when existing.

        Raises:
            (`FileExistsError`)
                If `overwrite` is False, when there already exists a path, it will raise Error.
        """
        self.trainer.save(path, overwrite)


class PyOperator(Operator):
    """
    Python function operators, no machine learning frameworks involved.
    """

    def __init__(self):
        super().__init__()
        pass

    @property
    def flag(self):
        return OperatorFlag.EMPTYFLAG



# Copyright (c) OpenMMLab. All rights reserved.
from ..utils import Registry, build_from_cfg


def build_model_from_cfg(cfg, registry, default_args=None):
    return build_from_cfg(cfg, registry, default_args)


MODELS = Registry('model', build_func=build_model_from_cfg)



"""Operator Loader."""
from collections import namedtuple
import os
import sys
import re
import time
import ctypes
import json
import warnings
import portalocker
from ..internal.edict import edict
from ..func import CFuncDef, bind, get_func_idcode, get_idcode_hash
from ..building.build import source_to_so_ctx, get_virtual_dirname, build_context, file_is_changed, ENV_PATH
from ..utils import get_git_hash, makedirs
from ..internal.dtype import DType, CStruct, TemplateType, CTYPENAME2CTYPE
from ..glue.common import CSTRUCT_CONSTRUCTOR
from ..glue.backend import get_glue_modules
from .gen_code import get_gen_rel_code
import antgo

gen_code = get_gen_rel_code(os.path.dirname(__file__))


if sys.version_info[0] >= 3:
    import importlib.util

    def load_module(name, pathname):
        """Load Module.

        Paramters
        ---------
        name: str
            the name of module.
        pathname:
            the name of path.

        Returns
        -------
        Module
        """
        spec = importlib.util.spec_from_file_location(name, pathname)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module
else:
    import imp

    def load_module(name, pathname):
        """Load Module.

        Paramters
        ---------
        name: str
            the name of module.
        pathname:
            the name of path.

        Returns
        -------
        Module
        """
        module = imp.load_source(name, pathname)
        return module


def _get_func_head_reg(name):
    """Get a pattern object for CFunction Head.

    Paramters
    ---------
    name: str
        Function name.

    Returns
    -------
    A pattern object
    """
    return re.compile(r'^\s*{}\s*(.*)'.format(name))


MOBULA_KERNEL_REG = _get_func_head_reg('EAGLEEYE_(KERNEL|FUNC)')

FUNC_REG = re.compile(
    r'^\s*(.*?)\s*\((.*?)\)(?:.*?)*')
CPP_TEMPLATE_REG = re.compile(r'^\s*template\s*\<(.*?)\>\s*')


def _get_template_decl(code):
    match = CPP_TEMPLATE_REG.search(code)
    if match is None:
        return None
    blocks = match.groups()[0].split(',')
    templates = []
    for block in blocks:
        block_sp = block.split()
        dtype, dname = block_sp
        if dtype.strip() == 'typename':
            templates.append(dname.strip())
    return templates


def parse_parameter_decl(decl):
    """Parse the code of parameter declaration

    Parameters
    ----------
    decl : str
        The C++ code of parameter declaration

    Returns
    -------
    Tuple
        (DType Instance,  variable name)
    """
    num_star = decl.count('*')
    assert num_star <= 1,\
        Exception('Only support pass-by-value or pass-by-1-level-pointer, \
            Error declaration: {}'.format(decl))
    is_pointer = num_star > 0
    if is_pointer:
        decl = decl.replace('*', '')
    decl = decl.strip()
    if decl.startswith('const '):
        is_const = True
        decl = decl[len('const '):]
    else:
        is_const = False
    decl_sp = decl.split(' ')

    # type_name and variable_name in C++ code
    type_name, var_name = decl_sp

    # void* func(...)
    if type_name == 'void':
        assert is_pointer
        return DType(ctypes.c_void_p, is_const=is_const), var_name

    # ctype func(...)
    ctype_name = 'c_{}'.format(type_name)
    if hasattr(ctypes, ctype_name):
        ctype = getattr(ctypes, ctype_name)
        if is_pointer:
            ctype = ctypes.POINTER(ctype)
        return DType(ctype, is_const=is_const), var_name

    if type_name in CSTRUCT_CONSTRUCTOR:
        info = CSTRUCT_CONSTRUCTOR[type_name]
        return CStruct(type_name, is_const=is_const, cstruct=info[0], constructor=info[1]), var_name

    # template type
    return TemplateType(tname=type_name, is_pointer=is_pointer, is_const=is_const), var_name


def parse_parameters_list(plist):
    """Parse the code of parameter declaration list

    Parameters
    ----------
    plist : str
        The code of parameter declaration list

    Returns
    -------
    rtn_type :
        The type of return value
    func_name : str
        function name
    pars_list: list
        [(DType|TemplateType, variable name), ...]
    """

    match = FUNC_REG.search(plist)
    head, plist = match.groups()
    head_split = re.split(r'\s+', head)
    plist_split = filter(lambda x: len(x) > 0, re.split(r'\s*,\s*', plist))
    func_name = head_split[-1]
    rtn_type = head_split[-2] if len(head_split) == 3 else None
    pars_list = list(map(parse_parameter_decl, plist_split))
    return rtn_type, func_name, pars_list


# runtime
FuncInfo = namedtuple('FuncInfo', ['func', 'cpp_info'])
CTX_FUNC_MAP = dict()  # CTX_FUNC_MAP[ctx][cpp_fname] -> FuncInfo


class CPPInfo:
    """The class of the C++ file's information.

    Parameters
    ----------
    cpp_fname: str
        the filename of C++ file.
    """

    def __init__(self, cpp_fname):
        self.cpp_fname = cpp_fname
        self.function_args = dict()
        self.dll = None

    def load_dll(self, dll_fname):
        """Load Dynamic-Link Library(*.so or *.dll).

        Parameters
        ----------
        dll_fname:
            The name of Dynamic-Link Library.
        """
        # keep reference
        self.dll = ctypes.CDLL(dll_fname)


def _build_lib(cpp_fname, code_buffer, ctx, target_name):
    # the virtual dirname of the source code
    cpp_path, cpp_basename = os.path.split(cpp_fname)
    create_time = time.strftime('%a %Y-%m-%d %H:%M:%S (%z)', time.localtime())
    git_hash = get_git_hash()
    extra_code = gen_code('./templates/wrapper.cpp')(
        cpp_fname=cpp_fname,
        git_hash=git_hash,
        create_time=create_time,
        inc_fname=os.path.abspath(cpp_fname),
        code=code_buffer)

    build_path_ctx = os.path.join(os.path.dirname(target_name),ctx)
    makedirs(build_path_ctx, exist_ok=True)

    # build so
    cpp_wrapper_fname = os.path.join(build_path_ctx,
                                     os.path.splitext(cpp_basename)[0] + '_wrapper.cpp')
    with open(cpp_wrapper_fname, 'w') as fout:
        fout.write(extra_code)

    # build lib
    srcs = [cpp_wrapper_fname]
    build_path = os.path.dirname(os.path.dirname(target_name))
    source_to_so_ctx(build_path, srcs, target_name, ctx)


def _dtype_to_tvm_value_type(dtype):
    if dtype.is_pointer:
        return 'v_handle'
    if 'int' in dtype.cname:
        return 'v_int64'
    return 'v_float64'


def _get_args_inst_mx(i, t):
    s = 'args.values[%d].%s' % (i, _dtype_to_tvm_value_type(t))
    if t.is_pointer:
        return '''
          static_cast<{dtype}>(
            static_cast<DLTensor*>({tv})->data)'''.format(dtype=t.cname, tv=s)
    else:
        s = '\n          ' + s
    return s


def _generate_kernel_code(func_idcode_hash, arg_types, arg_names, func_name):
    args_def = ', '.join(['{ctype} {name}'.format(
        ctype=dtype.cname,
        name=name
    ) for dtype, name in zip(arg_types, arg_names)])
    args_inst = ', '.join(arg_names)

    kernel_code = gen_code('./templates/kernel_code.cpp')(
        func_idcode_hash=func_idcode_hash,
        args_def=args_def,
        func_name=func_name,
        args_inst=args_inst)
    kernel_code += '\n'

    args_def_async_mx = ', '.join(['{ctype} {name}'.format(
        ctype='NDArrayHandle' if dtype.is_pointer else dtype.cname,
        name=name
    ) for dtype, name in zip(arg_types, arg_names)])

    using_async_mx = all(
        map(lambda dtype: 'void' not in dtype.cname, arg_types))
    if using_async_mx:
        args_inst_mx = [_get_args_inst_mx(i, t)
                        for i, t in enumerate(arg_types)]
        const_loc = []
        for i, dtype in enumerate(arg_types):
            if dtype.is_const and dtype.is_pointer:
                const_loc.append(i)
        num_const = len(const_loc)
        const_loc_code = 'nullptr' if num_const == 0 else 'std::array<int, %d>({%s}).data()' % (
            num_const, ','.join([str(u) for u in const_loc]))
        async_mx_code = gen_code('./templates/async_mx_code.cpp')(
            func_idcode_hash=func_idcode_hash,
            func_name=func_name,
            args_inst=args_inst,
            args_inst_mx=','.join(args_inst_mx),
            num_const=num_const,
            const_loc_code=const_loc_code,
            args_def_async_mx=args_def_async_mx,
        )
        async_mx_code += '\n'
        kernel_code += async_mx_code
    return kernel_code


def _generate_func_code(func_idcode_hash, rtn_type, arg_types, arg_names, func_name):
    if rtn_type is None:
        rtn_type = 'void'

    args_def = ', '.join(['{ctype} {name}'.format(
        ctype=dtype.cname,
        name=name
    ) for dtype, name in zip(arg_types, arg_names)])
    args_inst = ', '.join(arg_names)

    code = gen_code('./templates/func_code.cpp')(
        return_value=rtn_type,
        return_statement='' if rtn_type == 'void' else 'return',
        func_idcode_hash=func_idcode_hash,
        args_def=args_def,
        func_name=func_name,
        args_inst=args_inst,
    )

    return code


def _update_template_inst_map(idcode, template_functions, cfunc, arg_types):
    # template function
    func_name = cfunc.func_name
    func_idcode_hash = get_idcode_hash(idcode)
    # Check Template Type Mapping
    template_mapping = dict()
    for rtype, dtype in zip(arg_types, cfunc.arg_types):
        if not isinstance(dtype, TemplateType):
            continue
        tname = dtype.tname
        rtype = str(rtype).replace(
            'const', '').replace('*', '').strip()
        if tname in template_mapping:
            assert template_mapping[tname] == rtype,\
                Exception('Excepted template type {} instead of {}'.
                          format(template_mapping[tname], rtype))
        else:
            template_mapping[tname] = rtype
    assert len(template_mapping) == len(cfunc.template_list),\
        Exception('Template List: {}, mapping: {}'.
                  format(cfunc.template_list, template_mapping))

    template_inst = [template_mapping[tname]
                     for tname in cfunc.template_list]
    template_post = '<%s>' % (', '.join(template_inst)
                              ) if template_inst else ''
    rtn_type = cfunc.rtn_type
    if rtn_type in template_mapping:
        rtn_type = template_mapping[rtn_type]

    func_kind = cfunc.func_kind
    if func_kind == CFuncDef.KERNEL:
        code = _generate_kernel_code(func_idcode_hash, arg_types, cfunc.arg_names, '({}_kernel{})'.format(
            func_name, template_post))
    else:
        code = _generate_func_code(
            func_idcode_hash, rtn_type, arg_types, cfunc.arg_names, func_name + template_post)
    template_functions[idcode] = (code, rtn_type)


def _add_function(func_map, func_idcode, rtn_type, cpp_info, dll_fname):
    func_idcode_hash = get_idcode_hash(func_idcode)
    func = getattr(cpp_info.dll, func_idcode_hash, None)
    if func is None:
        functions = [name for name in dir(
            cpp_info.dll) if not name.startswith('_')]
        raise NameError('No function `{}` in DLL {}, current functions: {}'.format(
            func_idcode, dll_fname, functions))
    func.restype = CTYPENAME2CTYPE[rtn_type]

    old_func = func_map.get(func_idcode, None)
    if old_func is not None:
        if old_func.cpp_info.cpp_fname != cpp_info.cpp_fname:
            warnings.warn('The function `{}` in `{}` will be overridden by that in `{}`'.format(
                func_idcode, old_func.cpp_info.cpp_fname, cpp_info.cpp_fname))

    func_map[func_idcode] = FuncInfo(func=func, cpp_info=cpp_info)


class OpLoader:
    '''Import Operator Loader.
    It's actual to load the operator.

    Parameters
    ----------
    cfunc: CFuncDef
        The definition of function to call.
    arg_types: list of {DType|TemplateType}
        Argument declaration list.
    ctx: str
        Building context.
    cpp_info: CPPInfo
        Related to cfunc.

    Returns
    -------
    CTX_FUNC_MAP[ctx][fname][idcode] : FuncInfo
    '''

    def __init__(self, cfunc, arg_types, ctx, cpp_info):
        idcode = get_func_idcode(cfunc.func_name, arg_types)
        if ctx not in CTX_FUNC_MAP:
            CTX_FUNC_MAP[ctx] = dict()
        cpp_fname = cpp_info.cpp_fname
        if cpp_fname not in CTX_FUNC_MAP[ctx]:
            CTX_FUNC_MAP[ctx][cpp_fname] = dict()
        # func_map: dict mapping idcode to CFunction
        func_map = CTX_FUNC_MAP[ctx][cpp_fname]

        if idcode not in func_map:
            '''
            *load function* when one of the following conditions is True:
            1. idcode is not loaded
            2. loading the function with same function name but different cpp filename
            '''
            cpp_path, cpp_basename = os.path.split(cpp_fname)
            cpp_path = get_virtual_dirname(cpp_path)
            build_path = os.path.join(os.curdir, '.temp', 'op', cpp_path.split('/')[-1], 'build')

            use_template = bool(cfunc.template_list)
            makedirs(build_path, exist_ok=True)
            build_info_fname = os.path.join(
                build_path, os.path.splitext(cpp_basename)[0] + '.json')
            build_info_fs = open(build_info_fname, 'a+')
            portalocker.lock(build_info_fs, portalocker.LOCK_EX)
            build_info_fs.seek(0)
            js_data = build_info_fs.read()
            if js_data:
                map_data = json.loads(js_data)
            else:
                map_data = dict(version=antgo.__version__)
            del js_data

            # try to load the instance of template function
            # map_data is a dict which records build information
            if map_data.get('version') > antgo.__version__:
                portalocker.unlock(build_info_fs)
                raise Exception(
                    """Unsupported higher version %s of wrapper file (Current AntgoOP ver: %s) :-(.
Please update AntgoOP.""" % (map_data.get('version'), antgo.__version__))
            build_id = map_data.get('build_id', 0)
            is_old_version = map_data.get(
                'version') < antgo.__version__
            # load the information of template functions
            TEMPLATE_FUNCTION_NAME = 'functions'
            if is_old_version:
                template_functions = dict()
            else:
                template_functions = map_data.get(
                    TEMPLATE_FUNCTION_NAME, dict())

            so_prefix = os.path.join(build_path, os.path.splitext(cpp_basename)[0])
            # The filename of build target
            dll_fname_format = '{prefix}_{ctx}'.format(
                prefix=so_prefix, ctx=ctx) + '_{build_id}.so'
            dll_fname = dll_fname_format.format(build_id=build_id)

            file_changed = file_is_changed(cpp_fname)
            dll_existed = os.path.exists(dll_fname)
            func_existed = idcode in template_functions

            if file_changed or not dll_existed or not func_existed or is_old_version:
                # Rebuild DLL file
                try:
                    # try to remove old DLL file
                    os.remove(dll_fname)
                except:
                    pass
                if file_changed:
                    # clear template_functions since some functions may have been deleted or renamed after codefile is changed.
                    template_functions.clear()
                if file_changed or not func_existed:
                    '''
                    we increase `build_id` by 1 when one of the following conditions is True:
                    1. the cpp file has been changed
                    2. new idcode

                    When the cpp file is not changed, and idcode exists in template_functions,
                    `build_id` will be not changed.
                    '''
                    build_id += 1
                dll_fname = dll_fname_format.format(build_id=build_id)
                # build code
                if idcode not in template_functions:
                    _update_template_inst_map(
                        idcode, template_functions, cfunc, arg_types)
                # collects template instances code into code_buffer
                code_buffer = ''.join([v[0]
                                       for v in template_functions.values()])

                with build_context():
                    try:
                        _build_lib(cpp_fname, code_buffer, ctx, dll_fname)
                    except:
                        # if build fail, unlock the build info file
                        portalocker.unlock(build_info_fs)
                        raise
                # update template_functions
                map_data = dict(version=antgo.__version__,
                                build_id=build_id)
                map_data[TEMPLATE_FUNCTION_NAME] = template_functions
                # clear the old context and write json data
                build_info_fs.seek(0)
                build_info_fs.truncate()
                json.dump(map_data, build_info_fs)
                build_info_fs.flush()
                os.fsync(build_info_fs.fileno())
            portalocker.unlock(build_info_fs)

            # load all functions in the dll
            cpp_info.load_dll(dll_fname)

            # import all functions
            for func_idcode in template_functions.keys():
                _add_function(func_map,
                              func_idcode, template_functions[func_idcode][1], cpp_info, dll_fname)

        # 算子/函数
        self.func = func_map[idcode].func
        self.cpp_info = func_map[idcode].cpp_info
        self.idcode_hash = get_idcode_hash(idcode)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def clear(self, data):
        # 资源销毁函数
        getattr(self.cpp_info.dll, f'destroy_{str(data._obj.__class__.__name__).lower()}')(data)

    def get_async_func(self, glue_mod):
        async_name = getattr(glue_mod, 'async_name', None)
        if async_name is None:
            return None
        return glue_mod.get_async_func(self.cpp_info, self.idcode_hash)


def _get_functions_from_cpp(cpp_fname):
    unmatched_brackets = 0
    func_def = ''
    func_kind = ''
    func_started = False
    template_list = []
    cpp_info = CPPInfo(cpp_fname=cpp_fname)
    function_args = cpp_info.function_args
    for line in open(cpp_fname):
        if not func_started:
            current_template_list = _get_template_decl(line)
            if current_template_list is not None:
                template_list = current_template_list
            match = MOBULA_KERNEL_REG.search(line)
            if match is not None:
                func_def = ''
                func_kind_str = match.groups()[0]
                if func_kind_str == 'KERNEL':
                    func_kind = CFuncDef.KERNEL
                elif func_kind_str == 'FUNC':
                    func_kind = CFuncDef.FUNC
                else:
                    raise TypeError(
                        'Unknown kind of function: %s' % func_kind_str)
                func_started = True
        # In a declaration of a function
        if func_started:
            unmatched_brackets += line.count('(') - line.count(')')
            func_def += line
            if unmatched_brackets == 0:
                func_def = func_def.replace('\n', '').replace('\r', '')
                func_started = False
                rtn_type, kernel_name, par_list = parse_parameters_list(
                    func_def)
                # template name check
                template_set = set(template_list)
                assert len(template_set) == len(template_list),\
                    Exception('Duplicated template name in {}'.format(
                        ', '.join(template_list)))
                use_template = False
                for dtype, _ in par_list:
                    if isinstance(dtype, TemplateType):
                        assert dtype.tname in template_set,\
                            Exception(
                                "template name '{}' is not defined".format(dtype.tname))
                        use_template = True
                if not use_template:
                    template_list = []

                if func_kind == CFuncDef.KERNEL:
                    assert kernel_name.endswith('_kernel'),\
                        Exception('the postfix of a MOBULA_KERNEL name must be `_kernel`, \
                            e.g. addition_forward_kernel')
                    func_name = kernel_name[:-len('_kernel')]
                elif func_kind == CFuncDef.FUNC:
                    func_name = kernel_name
                else:
                    raise Exception(
                        'Unknown function kind: {}'.format(func_kind))

                # Arguments
                funcdef_args = edict(func_name=func_name,
                                     func_kind=func_kind,
                                     arg_names=[t[1] for t in par_list],
                                     arg_types=[t[0] for t in par_list],
                                     rtn_type=rtn_type,
                                     template_list=template_list,
                                     loader=OpLoader,
                                     loader_kwargs=dict(
                                         cpp_info=cpp_info,
                                     )
                                     )
                template_list = []
                function_args[func_name] = funcdef_args

    assert unmatched_brackets == 0,\
        Exception('# unmatched brackets: {}'.format(unmatched_brackets))

    # Load dynamic file
    functions = dict(
        (name, CFuncDef(**kwargs)) for name, kwargs in function_args.items())
    # Load dynamic function for MXNet
    return functions


def load(module_name, path=''):
    """Load Operator Module

    Parameters
    ----------
    module_name: str
        The name of Operator Module
    path: str
        The path of Operator Module [default = current path]
    """
    op_name = os.path.basename(module_name)
    if not path:
        # Find Operator Module in custom directory first
        custom_path = os.path.join(os.path.dirname(__file__), '../zoo')
        if os.path.exists(os.path.join(custom_path, op_name)):
            path = custom_path
    path = os.path.join(path, module_name)

    found = False
    cpp_fname = os.path.join(path, op_name + '.cpp')
    if os.path.exists(cpp_fname):
        found = True
        # Get functions
        functions = _get_functions_from_cpp(cpp_fname)
        bind(functions)
